"""
Notochord MIDI co-improviser server.
Notochord plays different instruments along with the player.
"""

# Authors:
#   Victor Shepardson
#   Intelligent Instruments Lab 2023

tui_doc = """Welcome to notochord homunculus.
You have 16 channels of MIDI, laid out in two rows of eight.
Each can be in one of three modes:
    input (-->02)
    follow (01->02)
    auto (02)
Input channels take MIDI input from the corresponding channel,
and send it to notochord. When using --thru, input is also copied
to the output.
    Follow channels act like a harmonizer; they query notochord for
a NoteOn whenever the followed channel has a NoteOn, and have the
corresponding NoteOff when the followed channel does.
    Auto channels are played autonomously by notochord.
Each channel can be muted independently of the others.
    Below the channel strips, there is a row of presets.
These can be edited in homunculus.toml; run `notochord files` to find 
it in your file explorer."""

# TODO: move soundfont / general MIDI stuff out of script

# TODO: make key bindings visibly click corresponding buttons
# TODO: make Mute a toggle but Reset a momentary

# TODO: color note by pitch class + register
# TODO: color instrument 1-128, MEL, DRUM, ANON, ANONDRUM
# TODO: color time
# TODO: id prediction as player / noto
# TODO: unify note log / prediction format
# TODO: grey out predictions when player or muted notochord

# TODO: controls display panel
# TODO: held notes display panel
# TODO: counts / weights display panel
# TODO: MIDI learn

import time
import shutil
from typing import Optional, Dict, Any
from numbers import Number
import math
import functools as ft
from pathlib import Path
from collections import defaultdict
from datetime import datetime

import numpy as np
import toml_file
import mido
from tqdm import tqdm
import joblib

import torch
torch.set_num_threads(1)

import iipyper, notochord
from notochord import Notochord, NotoPerformance
from iipyper import OSC, MIDI, run, Stopwatch, repeat, cleanup, TUI, profile, lock

from rich.panel import Panel
from rich.pretty import Pretty
from textual.reactive import reactive
from textual.widgets import Header, Footer, Static, Button, Log, RichLog, Label
from textual.screen import Screen, ModalScreen
from textual.containers import Grid

def now():
    return time.perf_counter()

def main(
    checkpoint="notochord-latest.ckpt", # Notochord checkpoint
    config:Dict[int,Dict[str,Any]]=None, # map MIDI channel : GM instrument
    preset:str=None,
    preset_file:Path=None,

    midi_prompt:Path=None,
    prompt_instruments:bool=True,

    initial_mute=False, # start with Notochord muted
    initial_query=False, # let Notochord start playing immediately

    midi_in:Optional[str]=None, # MIDI port for player input
    midi_out:Optional[str]=None, # MIDI port for Notochord output
    thru=False, # copy player input to output
    send_pc=False, # send program change messages
    dump_midi=False, # print all incoming MIDI
    suppress_midi_feedback=True,
    input_channel_map=None,

    balance_sample=False, # choose instruments which have played less recently
    n_recent=32, # number of recent note-on events to consider for above
    n_margin=8, # amount of 'slack' in the balance_sample calculation
    
    max_note_len=5, # in seconds, to auto-release stuck Notochord notes
    max_time=None, # max time between events
    nominal_time=True, #DEPRECATED

    osc_port=None, # if supplied, listen for OSC to set controls on this port
    osc_host='', # leave this as empty string to get all traffic on the port

    use_tui=True, # run textual UI
    predict_input=True, # forecasted next events can be for input (preserves model distribution, but can lead to Notochord deciding not to play)
    predict_follow=False,
    debug_query=False, # don't query notochord when there is no pending event.
    testing=False,
    estimated_query_latency=10e-3,
    estimated_feed_latency=10e-3,
    lateness_margin=100e-3, # when events are playing later than this, slow down
    soundfont=None,
    limit_input=None,
    thru_vel_offset=None,
    profiler=0,
    wipe_presets=False,
    ):
    """
    This a terminal app for using Notochord interactively with MIDI controllers and synthesizers. It allows combining both 'harmonizer' and 'improviser' features.
     
    Arguments to main can be given on the command line as flags, for example:

    `notochord homunculus --config "{1:{mode:auto, inst:1, mute:False}}" --initial-query --max-time 1`

    This says: play the grand piano autonomously on channel 1, start automatically, allow no more than 1 second between MIDI events. A different example:

    `notochord homunculus --config "{1:{mode:input, inst:1, mute:False}, 2:{mode:follow, inst:12, mute:False}}" --thru --send-pc`

    This says, take grand piano input on channel 1, and harmonize it with vibraphone on channel 2. Pass input through to the output, and also send program change messages to set the instruments on the synthesizer.

    You may also need to use the --midi-in or --midi-out flags to get MIDI to the right place.
    
    # MIDI Channels

    In homunculus, you have 16 MIDI channels. Each channel can be in one of three modes:

        * input (appearing like "-->01"), channel 1 comes from MIDI input channel 1.
        * follow (appearing like "01->02"), channel 2 plays whenever channel 1 plays.
        * auto (appearing like just "03"), channel 3 plays autonomously.

    Click the top section of each channel strip to cycle the mode.

    Each channel is also assigned a [General MIDI instrument](https://en.wikipedia.org/wiki/General_MIDI#Program_change_events). Each 'input' and 'auto' channel should have a unique General MIDI instrument, but 'follow' channels can be duplicates of others. If you try to assign duplicate instruments, they will automatically change to "anonymous" melodic or drum instruments, but still send the right program change messages when using --send-pc.

    Click the middle section of each channel strip to choose a new instrument.

    The bottom section of each channel strip allows muting individual voices.

    # Global controls

    Along the bottom, there are global query, sustain, mute and reset buttons. Query manually replaces the next pending note. Sustain stops all auto voices from playing without ending any open notes. Mute ends all open notes and stops auto voices. Reset ends all open notes, forgets all context and sets the Notochord model to its initial state.

    Args:
        checkpoint: path to notochord model checkpoint.

        config: 
            mapping from MIDI channels to voice specs.
            MIDI channels and General MIDI instruments are indexed from 1.
            see wikipedia.org/wiki/General_MIDI for values.
            There are 3 modes of voice, 'auto', 'follow' and 'input'. For example,
            ```
            {
                1:{
                    'mode':'input', 'inst':1
                }, # input grand piano on MIDI channel 1
                2:{
                    'mode':'follow', 'source':1, 'inst':1, 
                    'transpose':(-12,12)
                }, # harmonize the channel within 1 octave 1 with more piano
                3:{
                    'mode':'auto', 'inst':12, 'range':(36,72)
                }, # autonomous vibraphone voice in the MIDI pitch 36-72 range
                10:{
                    'mode':'auto', 'inst':129,
                }, # autonomous drums voice
                4:{
                    'mode':'follow', 'source':3, 'inst':10, 'range':(72,96)
                }, # harmonize channel 3 within upper registers of the glockenspiel
            }
            ```
            Notes:
            when two channels use the same instrument, one will be converted
            to an 'anonymous' instrument number (but it will still respect the 
            chosen instrument when using --send-pc)

        preset: 
            preset name (in preset file) to load config from
        preset_file: 
            path to a TOML file containing channel presets
            the default config file is `homunculus.toml`; 
            running `notochord files` will show its location.

        midi_prompt:
            path to a MIDI file to read in as a prompt
        prompt_instruments:
            if True, set unmuted instruments to those in the prompt file

        initial_mute: 
            start 'auto' voices muted so it won't play with input.
        initial_query: 
            query Notochord immediately,
            so 'auto' voices begin playing  without input.

        midi_in: 
            MIDI ports for input. 
            default is to use all input ports.
            can be comma-separated list of ports.
        midi_out: 
            MIDI ports for output. 
            default is to use only virtual 'From iipyper' port.
            can be comma-separated list of ports.
        thru: 
            if True, copy input MIDI to output ports.
            only makes sense if input and output ports are different.
        send_pc: 
            if True, send MIDI program change messages to set the General MIDI
            instrument on each channel according to player_config and noto_config.
            useful when using a General MIDI synthesizer like fluidsynth.
        dump_midi: 
            if True, print all incoming MIDI for debugging purposes

        balance_sample:
            choose 'auto' voices which have played less recently,
            ensures that all configured instruments will play.
        n_recent: 
            number of recent note-on events to consider for above
        n_margin: 
            controls the amount of slack in the balance_sample calculation

        max_note_len: 
            time in seconds after which to force-release sustained 'auto' notes.
        max_time: 
            maximum seconds between predicted events for 'auto' voices.
            default is the Notochord model's maximum (usually 10 seconds).
        nominal_time: 
            deprecated. when False, now sets lateness_margin to 0
        lateness_margin:
            when events are playing later than this (in seconds), slow down

        osc_port: 
            optional. if supplied, listen for OSC to set controls
        osc_host: 
            hostname or IP of OSC sender.
            leave this as empty string to get all traffic on the port

        use_tui: 
            run textual UI.
        predict_input: 
            forecasted next events can be for 'input' voices.
            generally should be True for manual input;
            use balance_sample to force 'auto' voices to play. 
            you might want it False if you have a very busy input.

        wipe_presets:
            if True, replaces your homunulus.toml with the defaults
            (may be useful after updating notochord)
    """
    if osc_port is not None:
        osc = OSC(osc_host, osc_port)
    midi = MIDI(midi_in, midi_out, suppress_feedback=suppress_midi_feedback)

    estimated_latency = estimated_feed_latency + estimated_query_latency

    if input_channel_map is None: input_channel_map = {}

    if soundfont is not None:
        # attempt to get instrument ranges from the soundfont
        # assumes first bank is used
        # not sure if entirely correct
        from sf2utils.sf2parse import Sf2File
        from sf2utils.generator import Sf2Gen

        with open(soundfont, 'rb') as file:
            soundfont = Sf2File(file)
        sf_presets = {
            p.preset:p
            for p in soundfont.presets 
            if hasattr(p,'bank') and p.bank==0}
        
        def _get_range(i):
            if i>127:
                return 0, 127
            lo, hi = 127,0
            for b in sf_presets[i-1].bags:
                if Sf2Gen.OPER_INSTRUMENT not in b.gens: 
                    continue
                if b.key_range is None: 
                    return 0, 127
                l,h = b.key_range
                lo = min(lo, l)
                hi = max(hi, h)
            assert lo<hi, (i-1,lo,hi)
            return lo, hi
        sf_inst_ranges = {i:_get_range(i) for i in range(1,129)}
        def get_range(i):
            return sf_inst_ranges.get(i, (0,127))
    else:
        def get_range(i):
            return 0,127  

    ### Textual UI
    tui = NotoTUI()
    print = notochord.print = iipyper.print = tui.print
    ###

    if not nominal_time:
        print('nominal_time is deprecated; nominal_time=False now sets lateness_margin to 0')
        lateness_margin = 0
    
    # make preset file if it doesn't exist
    cfg_dir = Notochord.user_data_dir()
    default_preset_file = cfg_dir / 'homunculus.toml'
    src_preset_file = Path(__file__).parent / 'homunculus.toml'
    if wipe_presets or not default_preset_file.exists():
        shutil.copy(src_preset_file, default_preset_file)

    ### presets and config
    try:    
        if preset_file is None:
            global_config = toml_file.Config(str(default_preset_file))
        else:
            global_config = toml_file.Config(str(preset_file))
        presets = global_config.get('preset', {})
        control_meta = global_config.get('control', {})
        action_meta = global_config.get('action', {})

    except Exception:
        print('WARNING: failed to load presets from file')
        global_config = {}
        presets = {}
        control_meta = []
        action_meta = []

    # store control values
    controls = {ctrl['name']:ctrl.get('value', None) for ctrl in control_meta}

    # defaults
    def default_config_channel(i):
        return {'mode':'auto', 'inst':257, 'mute':True, 'mono':False, 'source':max(1,i-1), 'note_shift':0}
    
    # convert MIDI channels to int
    # print(presets)
    for p in presets:
        p['channel'] = {
            int(k):{**default_config_channel(i), **v} 
            for i,(k,v) in enumerate(p['channel'].items(),1)}
    
    # load notochord model
    try:
        noto = Notochord.from_checkpoint(checkpoint)
        noto.eval()
        noto.feed(0,0,0,0) # smoke test
        noto.query()
        noto.reset()
    except Exception:
        print("""error loading notochord model""")
        raise

    ### this feeds all events from the prompt file to notochord
    if midi_prompt is None:
        initial_state = None
        config_ingest = None
    else:
        initial_state, mid_channel_inst = noto.prompt(midi_prompt)
        config_ingest = {
            c+1:{'inst':i, 'mode':'auto', 'mute':False} 
            for c,i in mid_channel_inst.items()}
        if not prompt_instruments:
            config_ingest = None

    # TODO: config from CLI > config from preset file > channel defaults
    #       warn if preset is overridden by config
    # TODO: what is sensible default behavior?
    #    for quickstart, it's good if it does something as soon as possible
    #    but for general use, it's good if it does nothing until you ask... 

    config_file = {}
    if isinstance(preset, str):
        for p in presets:
            if preset == p['name']:
                config_file = p['channel']
                if config is not None:
                    print('WARNING: `--config` overrides `--preset`')
                break
    elif len(presets):
        config_file = {**presets[0]['channel']}
        # preset = list(presets)[0]

    config_cli = {} if config is None else config

    config = {i:default_config_channel(i) for i in range(1,17)}
    for k,v in config_file.items():
        config[k].update(v)
    for k,v in config_cli.items():
        config[k].update(v)
    if config_ingest is not None:
        for k,v in config_ingest.items():
            config[k].update(v)

    # print(f'{config=}')

    def validate_config():
        assert all(
            v['source'] in config for v in config.values() if v['mode']=='follow'
            ), 'ERROR: no source given for follow voice'
        # TODO: check for follow cycles
    validate_config()

    # for c,v in config.items():
        # tui.set_inst(c, v['inst'])

    def mode_insts(t, allow_muted=True):
        if isinstance(t, str):
            t = t,
        # set of instruments with given mode(s)
        return {
            v['inst'] for v in config.values() 
            if v['mode'] in t and (allow_muted or not v['mute'])
            }
    def mode_chans(t):
        if isinstance(t, str):
            t = t,
        # list of channels with given mode
        return [k for k,v in config.items() if v['mode'] in t]
    def channel_inst(c):
        return config[c]['inst']
    def channel_insts():
        # list of channel,instrument pairs
        return [(c,channel_inst(c)) for c in config]
    def inst_ranges(insts):
        # instruments to sets of allowed MIDI numbers
        r = {}
        for v in config.values():
            i = v['inst']
            if i in insts:
                s = set(range(*v.get('range', get_range(i))))
                if i in r:
                    r[i] |= s
                else:
                    r[i] = s
        return r
    def auto_inst_channel(i):
        for k,v in config.items():
            if v['inst']==i:
                return k
        raise ValueError
    def channel_followers(chan):
        # return channel of all 'follow' voices with given source
        return [
            k for k,v in config.items() 
            if v['mode']=='follow' 
            and v.get('source', None)==chan]
      
    def dedup_inst(c, i):
        # change to anon if already in use
        def in_use():
            return any(
                c_other!=c 
                and config[c_other]['inst']==i 
                and config[c_other]['mode']!='follow'
                for c_other in config)
        if in_use():
            # print(i, config)
            i = noto.first_anon_like(i)
        while in_use():
            i = i+1
        return i

    def do_send_pc(c, i):
        # warn_inst(i)
        # assuming fluidsynth -o synth.midi-bank-select=mma
        if noto.is_drum(i):
            midi.control_change(channel=c-1, control=0, value=1)
        else:
            midi.control_change(channel=c-1, control=32, value=0)
            midi.control_change(channel=c-1, control=0, value=0)
        if noto.is_anon(i):
            program = 0
        else:
            # convert to 0-index
            program = (i-1) % 128
        midi.program_change(channel=c-1, program=program)

    for c,i in channel_insts():
        if send_pc:
            do_send_pc(c, i)
        config[c]['inst'] = dedup_inst(c, i)

    # simple class to hold pending event prediction
    class Prediction:
        def __init__(self):
            self.gate = not initial_mute
            self.clear()
        def clear(self):
            self.event = None
            self.last_event_time = None
            self.next_event_time = None
            self.lateness = 0
            tui.defer(prediction=None)
        def occurred(self):
            self.event = None
            self.last_event_time = self.next_event_time
        def set(self, event):
            if event.get('time') is None:
                event['time'] = self.time_since()
            self.event = event
            self.next_event_time = event['time'] + (
                self.last_event_time or now())
            tui.defer(prediction=pending.event)
            # print(f'{now()=} {self.next_event_time=}')
        def time_since(self):
            if self.last_event_time is None:
                return 0
            else:
                return now() - self.last_event_time
        def time_until(self):
            if self.next_event_time is None: 
                return float('inf')
            else:
                return self.next_event_time - now()
        def is_auto(self):
            if self.event is None:
                return False
            return self.event['inst'] in mode_insts('auto')
    pending = Prediction()    

    # tracks held notes, recently played instruments, etc
    history = NotoPerformance()

    follow_status = {'depth':0}

    def display_event(tag, memo, inst, pitch, vel, channel, **kw):
        """print an event to the terminal"""
        if tag is None:
            return
        now = str(datetime.now())[:-2]
        s = f'{now}   inst {inst:3d}   pitch {pitch:3d}   vel {vel:3d}   ch {channel:2d} {tag}'
        # s = f'{tag}:\t{inst=:4d}   {pitch=:4d}   {vel=:4d}   {channel=:3d}'
        if memo is not None:
            s += f' ({memo})'
        tui.defer(note=s)

    def send_midi(note, velocity, channel):
        kind = 'note_on' if velocity > 0 else 'note_off'
        cfg = config[channel]
        # get channel note map
        if 'note_map' in cfg:
            note_map = cfg['note_map']
        else:
            note_map = {}
        if note in note_map:
            note = note_map[note]
        elif 'note_shift' in cfg:
            note = note + cfg['note_shift']

        port = cfg.get('port', None)

        midi.send(kind, note=note, velocity=velocity, channel=channel-1, port=port)

    def play_event(
            channel, 
            parent=None, # parent note as (channel, inst, pitch)
            feed=True, 
            send=True, 
            tag=None, memo=None):
        """realize an event as MIDI, terminal display, and Notochord update"""
        event = pending.event
        if event is None:
            print("WARNING: play_event on null event")
            return
        time_until = pending.time_until()
        pending.lateness = -time_until # set at the time of playing
        if time_until < -0.1:
            # print(f'late {-time_until} {pending.last_event_time=} {pending.next_event_time=} {event.get("time")=} {now()=}')
            print(f'late {-time_until} at {now()=}')
        pending.occurred()
        # normalize values
        vel = event['vel'] = math.ceil(event['vel'])

        # send out as MIDI
        if send:
            with profile('\tmidi.send', print=print, enable=profiler>1):
                send_midi(event['pitch'], vel, channel)

        # print
        with profile('\tdisplay_event', print=print, enable=profiler>1):
            display_event(
                tag, memo=memo, channel=channel, **event)

        if feed:
            # feed to NotoPerformance
            # put a stopwatch in the held_note_data field for tracking note length
            with profile('\thistory.feed', print=print, enable=profiler>1):
                history.feed(held_note_data={
                    'duration':Stopwatch(),
                    'parent':parent
                    }, channel=channel, **event)
            # feed to model
            with profile('\tnoto.feed', print=print, enable=profiler>1):
                noto.feed(**event)

        follow_status['depth'] += 1
        with profile('\t'*follow_status['depth']+'follow.event', print=print, enable=profiler>1):
            # print(f'{follow_status=}')
            follow_event(event, channel)
            follow_status['depth'] -= 1
            # print(f'{follow_status=}')

    def follow_event(source_event, source_channel):
        source_vel = source_event['vel']
        source_pitch = source_event['pitch']
        source_inst = source_event['inst']
        source_k = (source_channel, source_inst, source_pitch)

        dt = 0

        if source_vel > 0:
            # NoteOn
            for noto_channel in channel_followers(source_channel):
                cfg = config[noto_channel]
                
                if cfg.get('mute', False): continue

                noto_inst = cfg['inst']
                min_x, max_x = cfg.get('transpose', (-128,128))
                lo, hi = cfg.get('range', (0,127))

                already_playing = {
                    note.pitch for note in history.notes if noto_inst==note.inst}
                # print(f'{already_playing=}')

                pitch_range = range(
                    max(lo,source_pitch+min_x), min(hi, source_pitch+max_x+1))
                pitches = (
                    set(pitch_range) - {source_pitch} - already_playing
                )

                if len(pitches)==0:
                    # edge case: no possible pitch
                    print(f'skipping {noto_channel=}, no pitches available')
                    print(f'{pitch_range} minus {{source_pitch}} minus {already_playing}')
                    continue
                elif len(pitches)==1:
                    # edge case: there is exactly one possible pitch
                    h = dict(
                        inst=noto_inst, pitch=list(pitches)[0], 
                        time=0, vel=source_vel)
                else:
                    # notochord chooses pitch
                    pending.set(noto.query(
                        next_inst=noto_inst, next_time=dt, next_vel=source_vel,
                        include_pitch=pitches))
                    
                play_event(
                    noto_channel, 
                    parent=source_k, tag='NOTO', memo='follow')
        # NoteOff
        else:
            # print(f'{history.note_data=}')
            dependents = [
                noto_k # chan, inst, pitch
                for noto_k, note_data
                in history.note_data.items()
                if note_data['parent']==source_k
            ]

            for noto_channel, noto_inst, noto_pitch in dependents:
                pending.set(dict(
                    inst=noto_inst, pitch=noto_pitch, time=dt, vel=0))
                play_event(noto_channel,tag='NOTO', memo='follow')

    # @lock
    def noto_reset():
        """reset Notochord and end all of its held notes"""
        print('RESET')

        # cancel pending predictions
        pending.clear()
        
        # end Notochord held notes
        # for (chan,inst,pitch) in history.note_triples:
        for note in history.notes:
            if note.inst in mode_insts('auto'):
                pending.set(dict(
                    inst=note.inst, pitch=note.pitch, vel=0))
                play_event(
                    channel=note.chan, 
                    feed=False, # skip feeding Notochord since we are resetting it
                    tag='NOTO', memo='reset')
                
        # reset notochord state
        noto.reset(state=initial_state)
        # reset history
        history.push()

        # TODO: feed note-ons from any held input/follower notes?

        # query the fresh notochord for a new prediction
        if pending.gate:
            auto_query(immediate=True)

    # @lock
    def noto_mute(sustain=False):
        tui.query_one('#mute').label = 'UNMUTE' if pending.gate else 'MUTE'
        # if sustain:
        tui.query_one('#sustain').label = 'END SUSTAIN' if pending.gate else 'SUSTAIN'

        pending.gate = not pending.gate

        if sustain:
            print('END SUSTAIN' if pending.gate else 'SUSTAIN')
        else:
            print('UNMUTE' if pending.gate else 'MUTE')
        # if unmuting, we're done
        if pending.gate:
            auto_query()
            # if sustain:
                # auto_query()
            return
        # cancel pending predictions
        pending.clear()

        if sustain:
            return
        
        # end+feed all held notes
        # for (chan,inst,pitch) in history.note_triples:
        for note in history.notes:
            if note.chan in mode_chans('auto'):
                pending.set(dict(
                    inst=note.inst, pitch=note.pitch, vel=0))
                play_event(
                    channel=note.chan, tag='AUTO', memo='mute')

    # query Notochord for a new next event
    # @lock
    def auto_query(
            predict_input=predict_input, 
            predict_follow=predict_follow,
            immediate=False):
        # check for stuck notes
        # and prioritize ending those
        for (_, inst, pitch), note_data in history.note_data.items():
            dur = note_data['duration'].read()
            if (
                inst in mode_insts('auto') 
                and dur > max_note_len*(.1+controls.get('steer_duration', 1))
                ):
                # query for the end of a note with flexible timing
                # with profile('query', print=print, enable=profiler):
                t = pending.time_since()
                mt = max(t, min(max_time or np.inf, t+0.2))
                pending.set(noto.query(
                    next_inst=inst, next_pitch=pitch,
                    next_vel=0, min_time=t, max_time=mt))
                print(f'END STUCK NOTE {inst=},{pitch=}')
                return

        # all_insts = mode_insts(('auto', 'input', 'follow'), allow_muted=True)
        # counts = history.inst_counts(n=n_recent, insts=all_insts)
        counts = defaultdict(int)
        for i,c in history.inst_counts(n=n_recent).items():
            counts[i] = c
        # print(f'{counts=}')

        # if using nominal time,
        # *subtract* estimated feed latency from min_time; unless immediately 
        # querying following a previous event, in which case let min_time be 0
        # if using actual time, *add* estimated query latency
        if immediate:
            min_time = 0
        else:
            time_offset = -estimated_feed_latency
            min_time = pending.time_since() + time_offset
        # print(f'{immediate=} {min_time=}')

        if pending.lateness > lateness_margin:
            backoff = pending.time_since() + pending.lateness/2
            if backoff > min_time:
                min_time = backoff
                print(f'set {min_time=} due to lateness')

        # if max_time < min_time, exclude input instruments
        input_late = max_time is not None and max_time < min_time

        inst_modes = ['auto']
        if predict_follow:
            inst_modes.append('follow')
        if predict_input and not input_late:
            inst_modes.append('input')
        allowed_insts = mode_insts(inst_modes, allow_muted=False)

        # assert len(allowed_insts - all_insts)==0, (allowed_insts, all_insts)

        # held_notes = history.held_inst_pitch_map(all_insts)
        held_notes = history.held_inst_pitch_map()
        # print(f'{held_notes=}')

        steer_time = 1-controls.get('steer_rate', 0.5)
        steer_pitch = controls.get('steer_pitch', 0.5)
        steer_density = controls.get('steer_density', 0.5)

        rhythm_temp = controls.get('rhythm_temp', 1)
        timing_temp = controls.get('timing_temp', 1)
        
        tqt = (max(0,steer_time-0.5), min(1, steer_time+0.5))
        tqp = (max(0,steer_pitch-0.5), min(1, steer_pitch+0.5))

        # idea: maintain an 'instrument presence' quantity
        # incorporating time since / number of notes playing
        # ideally this would distinguish sustained from percussive instruments too
        
        # balance_sample: note-ons only from instruments which have played less
        inst_weights = None
        if balance_sample:
            inst_weights = {}
            mc = max(counts.values()) if len(counts) else 0
            for i in allowed_insts:
                # don't upweight player controlled instruments though
                if i in mode_insts('auto'):
                    inst_weights[i] = np.exp(max(0, mc - counts[i] - n_margin))
                else:
                    inst_weights[i] = 1.

        # print(f'{inst_weights=}')

        # VTIP is better for time interventions,
        # VIPT is better for instrument interventions
        if min_time > estimated_latency or abs(steer_time-0.5) > abs(steer_pitch-0.5):
            # print('VTIP')
            query_method = noto.query_vtip
        else:
            # print('VIPT')
            query_method = noto.query_vipt

        # print(f'considering {insts} for note_on')
        # use only currently selected instruments
        inst_pitch_map = inst_ranges(allowed_insts)
        note_on_map = {
            i: set(inst_pitch_map[i])-set(held_notes[i]) # exclude held notes
            for i in allowed_insts#allowed_insts
            if i in inst_pitch_map
        }
        # use any instruments which are currently holding notes
        note_off_map = {
            i: set(held_notes[i]) # only held notes
            for i in allowed_insts
            if i in held_notes
        }

        # print(note_on_map, note_off_map)

        max_t = None if max_time is None else max(max_time, min_time+0.2)

        try:
            with profile('\tquery_method', print=print, enable=profiler>1):
                pending.set(query_method(
                    note_on_map, note_off_map,
                    min_time=min_time, max_time=max_t,
                    truncate_quantile_time=tqt,
                    truncate_quantile_pitch=tqp,
                    rhythm_temp=rhythm_temp,
                    timing_temp=timing_temp,
                    steer_density=steer_density,
                    inst_weights=inst_weights,
                    no_steer=mode_insts(('input','follow'), allow_muted=False),
                ))
        except Exception as e:
            print(f'WARNING: query failed. {allowed_insts=} {note_on_map=} {note_off_map=}')
            print(e.args)
            print(repr(e.__traceback__))
            pending.clear()

    #### MIDI handling

    # print all incoming MIDI for debugging
    if dump_midi:
        @midi.handle
        def _(msg):
            print(msg)

    # @midi.handle(type='program_change')
    # def _(msg):
        # """
        # Program change events set GM instruments on the corresponding channel
        # """
        # raise NotImplementedError

    @midi.handle(type='pitchwheel')
    def _(msg):
        """
        pitchwheel affects steer_pitch
        """
        controls['steer_pitch'] = (msg.pitch+8192)/16384
        # print(controls)

    # very basic CC handling for controls
    control_cc = {}
    control_osc = {}
    for ctrl in control_meta:
        name = ctrl['name']
        ccs = ctrl.get('control_change', [])
        if isinstance(ccs, Number) or isinstance(ccs, str):
            ccs = (ccs,)
        for cc in ccs:
            control_cc[cc] = ctrl
        control_osc[f'/notochord/homunculus/{name}'] = ctrl
    action_cc = {}
    action_osc = {}
    for act in action_meta:
        name = act['name']
        ccs = act.get('control_change', [])
        if isinstance(ccs, Number) or isinstance(ccs, str):
            ccs = (ccs,)
        for cc in ccs:
            action_cc[cc] = act
        action_osc[f'/notochord/homunculus/{name}'] = act

    def dispatch_action(k):   
        if k=='reset':
            noto_reset()
        elif k=='query':
            auto_query()
        elif k=='mute':
            noto_mute()
        else:
            print(f'WARNING: action "{k}" not recognized')

    @midi.handle(type='control_change')
    def _(msg):
        """
        these are global controls listening on all channels: 
        CC 01: steer pitch (>64 higher pitches, <64 lower)
        CC 02: steer density (>64 more simultaneous notes, <64 fewer)
        CC 03: steer rate (>64 more events, <64 fewer)
        """

        if msg.control in control_cc:
            ctrl = control_cc[msg.control]
            name = ctrl['name']
            lo, hi = ctrl.get('range', (0,1))
            controls[name] = msg.value/127 * (hi-lo) + lo
            print(f"{name}={controls[name]}")

        if msg.control in action_cc:
            k = action_cc[msg.control]['name']
            dispatch_action(k)

    if osc_port is not None:
        @osc.handle('/notochord/homunculus/*')
        def _(route, *a):
            # print('OSC:', route, *a)
            if route in control_osc:
                ctrl = control_osc[route]
                name = ctrl['name']
                assert len(a)==1
                arg = a[0]
                assert isinstance(arg, Number)
                lo, hi = ctrl.get('range', (0,1))   
                controls[name] = min(hi, max(0, arg))
                print(f"{name}={controls[name]}")

            if route in action_osc:
                k = action_osc[route]['name']
                dispatch_action(k)

    input_sw = Stopwatch()
    dropped = set()# (channel, pitch)
    input_dts = []
    @midi.handle(type=('note_on', 'note_off'))
    def _(msg):
        """
        MIDI NoteOn and NoteOff events affect input channels
        e.g. a channel displaying -->01 will listen to note events on channel 1
        a channel displaying 02->03 will follow note events on channel 2
        """
        channel = msg.channel + 1
        # convert from 0-index
        channel = input_channel_map.get(channel, channel)
        cfg = config[channel]

        if channel not in mode_chans('input'):
            print(f'WARNING: ignoring MIDI {msg} on non-input channel')
            return
        
        if cfg['mute']:
            print(f'WARNING: ignoring MIDI {msg} on muted channel')
            return
        
        # port = cfg.get('port', None)
        # if thru:
        #     if msg.velocity>0 and msg.type=='note_on' and thru_vel_offset is not None:
        #         vel = max(1, int(msg.velocity*thru_vel_offset))
        #         midi.send(msg.type, note=msg.note, channel=channel+1, velocity=vel, port=port)
        #     else:
        #         midi.send(msg, port=port, channel=channel+1)

        inst = channel_inst(channel)
        pitch = msg.note
        if 'note_shift' in cfg:
            pitch = pitch + cfg['note_shift']
        vel = msg.velocity if msg.type=='note_on' else 0
        
        dt = input_sw.punch()
        # print(f'EVENT {dt=} {msg}')
        if len(input_dts) >= 10:
            input_dts.pop(0)
        input_dts.append(dt)
        input_dens = len(input_dts) / sum(input_dts)
        # TODO: 
        # want to drop input when event density is high,
        # not just dt is short
        k = (channel,pitch)
        if vel==0 and k in dropped:
            dropped.remove(k)
            print(f'WARNING: ignoring rate-limited input')
            return
        if vel>0 and limit_input and input_dens>limit_input:
            print(f'WARNING: ignoring rate-limited input {input_dens=}')
            dropped.add(k)
            return 

        # feed event to Notochord
        # with profile('feed', print=print, enable=profiler):
        # with lock:
        pending.set({'inst':inst, 'pitch':pitch, 'vel':vel})
        play_event(channel=channel, send=thru, tag='PLAYER')
        # query for new prediction
        auto_query(immediate=True)

        # send a MIDI reply for latency testing purposes:
        # if testing: midi.cc(control=3, value=msg.note, channel=15)

    def auto_event():
        # print('auto_event')
        # 'auto' event happens:
        event = pending.event
        inst, pitch, vel = event['inst'], event['pitch'], math.ceil(event['vel'])
        chan = auto_inst_channel(inst)

        # note on which is already playing or note off which is not

        if (vel>0) == ((inst, pitch) in history.note_pairs): 
            print(f're-query for invalid {vel=}, {inst=}, {pitch=}')
            auto_query()
            return
        
        play_event(channel=chan, tag='NOTO')

    @repeat(lock=True)
    def _():
        """Loop, checking if predicted next event happens"""
        # check if current prediction has passed
        time_until = pending.time_until()
        # print(f'''
        #     {follow_status["depth"]=},
        #     {not testing=},
        #     {pending.gate=},
        #     {time_until=}''')
        if (
            follow_status['depth']==0 and
            not testing and
            pending.gate and
            pending.event and
            # stopwatch.read() > get_dt()
            time_until <= 0
            ):
            # if so, check if it is a notochord-controlled instrument
            if pending.is_auto():
                # prediction happens
                with profile('auto_event', print=print, enable=profiler):
                    auto_event()
                immediate = True
            else:
                immediate = False
            # query for new prediction
            # if get_dt() < noto.max_dt and not debug_query:
            if not debug_query:
                with profile('auto_query', print=print, enable=profiler):
                    auto_query(immediate=immediate)
        # adaptive time resolution here -- yield to other threads when 
        # next event is not expected, but time precisely when next event
        # is imminent
        wait = pending.time_until() if pending.is_auto() else 10e-3
        r = max(0, min(10e-3, wait))
        # print('wait', r)
        return r

    @cleanup
    def _():
        """end any remaining notes"""
        # print(f'cleanup: {notes=}')
        for note in history.notes:
            if note.inst in mode_insts(('auto', 'follow')):
                midi.note_off(note=note.pitch, velocity=0, channel=note.chan-1)
                # midi.note_on(note=note.pitch, velocity=0, channel=note.chan-1)

    ### update_* keeps the UI in sync with the state

    def update_config():
        # print(config)
        for c,v in config.items():
            tui.set_channel(c, v)

    def update_presets():
        for k,p in enumerate(presets):
            tui.set_preset(k, p.get('name'))

    @tui.on
    def mount():
        update_config()
        update_presets()
        print('MIDI handling:')
        print(midi.get_docs())
        if osc_port is not None:
            print('OSC handling:')
            print(osc.get_docs())
        print(tui_doc)
        print('For more detailed documentation, see:')
        print('https://intelligent-instruments-lab.github.io/notochord/reference/notochord/app/homunculus/')
        print('or run `notochord homunculus --help`')
        print('to exit, use CTRL+C')

        if initial_query:
            auto_query(predict_input=False, predict_follow=False)


    ### set_* does whatever necessary to change channel properties
    ### calls update_config() to keep the UI in sync

    def set_mode(c, m, update=True):
        if c in config:
            prev_m = config[c]['mode']
        else:
            prev_m = None
        if m==prev_m:
            return
        
        if m=='follow':
            if 'source' not in config[c]:
                print('WARNING: follower without a source, setting to 1')
                config[c]['source'] = 1
        
        config[c]['mode'] = m
        print(f'set channel {c} from {prev_m} to {m} mode')

        ### NOTE changing a channel with followers to input causes stuck notes

        if prev_m=='follow':
            # emancipate held notes
            for (dep_c,_,_), note_data in history.note_data.items():
                if dep_c==c:
                    note_data['parent'] = None

        if prev_m=='auto':
            # release held notes
            for note in history.notes:
                if note.chan==c:
                    pending.set(dict(inst=note.inst, pitch=note.pitch, vel=0))
                    play_event(
                        channel=note.chan, 
                        tag='NOTO', memo='mode change')

        if update:
            update_config()


    def set_inst(c, i, update=True):
        print(f'SET INSTRUMENT {i}')
        if c in config:
            prev_i = config[c]['inst']
        else:
            prev_i = None
        if prev_i==i:
            # print('SAME INSTRUMENT')
            return

        # for (chan,inst,pitch) in history.note_triples:
        for note in history.notes:
            if note.chan==c and config[note.chan]['mode']!='input':
                pending.set(dict(inst=note.inst, pitch=note.pitch, vel=0))
                play_event(
                    channel=note.chan, 
                    tag='NOTO', memo='change instrument')
        # send pc if appropriate
        if send_pc:
            do_send_pc(c, i)

        i = dedup_inst(c, i)
        
        # then set config
        config[c]['inst'] = i
        # and call:
        if update:
            update_config()

    @lock
    def set_mute(c, b, update=True):
        config[c]['mute'] = b
        if b:
            print(f'mute channel {c}')
            # release held notes
            # for (chan,inst,pitch) in history.note_triples:
            ended_any = False
            for note in history.notes:
                print(f'held {note=}')
                if note.chan==c and config[c]['mode']!='input':
                    pending.set(dict(inst=note.inst, pitch=note.pitch, vel=0))
                    play_event(
                        channel=note.chan, 
                        tag='NOTO', memo='mute channel')
                    ended_any = True
            if pending.gate:
                auto_query(immediate=ended_any)
        else:
            print(f'unmute channel {c}')
            if pending.gate:
                auto_query()

        if update:
            update_config()


    ### action_* runs on key/button press;
    ### invokes cycler / picker logic and calls set_*

    # this is pretty awful
    # need a better way to reconcile iipyper and textual here
    def action_mode(c):
        if c not in config: return
        # TODO: mode picker
        if config[c]['mode'] == 'auto':
            set_mode(c, 'input')
        elif config[c]['mode'] == 'input' and config[c]['source']!=c:
            # TODO: source picker for follow
            set_mode(c, 'follow')
        else:
            set_mode(c, 'auto')

    def action_inst(c):
        print(f'inst channel {c}')
        tui.push_screen(InstrumentGroupSelect(c))

    def action_mute(c):
        if i not in config: return
        set_mute(c, not config[c].get('mute', False))

    def action_preset(p):
        if p >= len(presets): 
            return
        preset = presets[p]['channel']
        print(f'load preset: {p}')
        for c in range(1,17):    
            if c not in config:
                config[c] = default_config_channel(c)
            if c not in preset:
                set_mute(c, True, update=False)
            else:
                # note config should *not* be updated before calling set_* 
                v = {**preset[c]}
                # v = {**config[c]}
                # v.update(preset[c])
                set_mode(c, v.pop('mode'), update=False)
                set_inst(c, v.pop('inst'), update=False)
                set_mute(c, v.pop('mute'), update=False)
                config[c].update(v)
        update_config()

    ### set actions which have an index argument
    ### TODO move this logic into @tui.set_action

    for i in range(1,17):
        setattr(tui, f'action_mode_{i}', ft.partial(action_mode, i))
        setattr(tui, f'action_inst_{i}', ft.partial(action_inst, i))
        setattr(tui, f'action_mute_{i}', ft.partial(action_mute, i))

    for i in range(10):
        setattr(tui, f'action_preset_{i}', ft.partial(action_preset, i))

    ### additional key/button actions

    @tui.set_action
    def mute():
        noto_mute()

    @tui.set_action
    def sustain():
        noto_mute(sustain=True)
    
    @tui.set_action
    def reset():
        noto_reset()
    
    @tui.set_action
    def query():
        auto_query()

    ### TUI classes which close over variables defined in main

    class Instrument(Button):
        """button which picks an instrument"""
        def __init__(self, c, i):
            super().__init__(inst_label(i))
            self.channel = c
            self.inst = i
        def on_button_pressed(self, event: Button.Pressed):
            self.app.pop_screen()
            self.app.pop_screen()
            set_inst(self.channel, self.inst)

    class InstrumentGroup(Button):
        """button which picks an instrument group"""
        def __init__(self, text, c, g):
            super().__init__(text)
            self.channel = c
            self.group = g
        def on_button_pressed(self, event: Button.Pressed):
            # show inst buttons
            tui.push_screen(InstrumentSelect(self.channel, self.group))

    class InstrumentSelect(ModalScreen):
        """Screen with instruments"""
        def __init__(self, c, g):
            super().__init__()
            self.channel = c
            self.group = g

        def compose(self):
            yield Grid(
                *(
                    Instrument(self.channel, i)
                    for i in gm_groups[self.group][1]
                ), id="dialog",
            )

    class InstrumentGroupSelect(ModalScreen):
        """Screen with instrument groups"""
        # TODO: add other features to this screen -- transpose, range, etc?
        def __init__(self, c):
            super().__init__()
            self.channel = c

        def compose(self):
            yield Grid(
                *(
                    InstrumentGroup(s, self.channel, g)
                    for g,(s,_) in enumerate(gm_groups)
                ), id="dialog",
            )

    if use_tui:
        tui.run()
    elif initial_query:
        auto_query(predict_input=False, predict_follow=False)

### def TUI components ###
class NotoLog(Log):
    value = reactive('')
    def watch_value(self, time: float) -> None:
        self.write_line(self.value)

class NotoPrediction(Static):
    value = reactive(None)
    def watch_value(self, time: float) -> None:
        evt = self.value
        if evt is None:
            s = ''
        else:
            s = f"\tinstrument: {evt['inst']:3d}    pitch: {evt['pitch']:3d}    time: {int(evt['time']*1000):4d} ms    velocity:{int(evt['vel']):3d}"
        self.update(Panel(s, title='prediction'))

class Mixer(Static):
    def compose(self):
        for i in range(1,17):
            yield MixerButtons(i, id=f"mixer_{i}")

def preset_id(i):
    return f"preset_{i}"
def mode_id(i):
    return f"mode_{i}"
def inst_id(i):
    return f"inst_{i}"
def mute_id(i):
    return f"mute_{i}"
class MixerButtons(Static):
    def __init__(self, idx, **kw):
        self.idx = idx
        super().__init__(**kw)

    def compose(self):
        yield Button(
            f"{self.idx:02d}", 
            id=mode_id(self.idx),
            classes="cmode"
            )
        yield Button(
            f"--- \n-----\n-----", 
            id=inst_id(self.idx),
            classes="cinst"
            )
        yield Button(
            f"MUTE", 
            id=mute_id(self.idx),
            classes="cmute"
            )
        
    def on_mount(self) -> None:
        lines = [
            "cycle channel mode:",
            f"-->{self.idx:02d}   input from MIDI channel {self.idx}"
        ]
        if self.idx > 0: lines.append(
            f"{self.idx-1:02d}->{self.idx:02d}  follow channel {self.idx-1}")
        lines.append(
            f"{self.idx:02d}      notochord plays autonomously"
        )
        self.query_one("#"+mode_id(self.idx)).tooltip = '\n'.join(lines)
        self.query_one("#"+inst_id(self.idx)).tooltip = "open instrument picker"
        self.query_one("#"+mute_id(self.idx)).tooltip = f"mute channel {self.idx}"
        
class NotoPresets(Static):
    n_presets = 10
    def compose(self):
        for i in range(NotoPresets.n_presets):
            yield Button('---', id=preset_id(i))

    def on_mount(self) -> None:
        for i in range(NotoPresets.n_presets):
            self.query_one("#"+preset_id(i)).tooltip = f"load preset {i}"

class NotoControl(Static):
    def compose(self):
        # yield NotoToggle()
        # yield Checkbox("Mute", id="mute")
        yield Button("Mute", id="mute", variant="error")
        yield Button("Sustain", id="sustain", variant="primary")
        yield Button("Query", id="query")
        yield Button("Reset", id="reset", variant="warning")

    def on_mount(self) -> None:
        self.query_one("#mute").tooltip = "master mute notochord"
        self.query_one("#sustain").tooltip = "master sustain -- prevent any NoteOffs from notochord"
        self.query_one("#query").tooltip = "manually query a new event"
        self.query_one("#reset").tooltip = "reset notochord"

class NotoTUI(TUI):
    CSS_PATH = 'homunculus.css'

    BINDINGS = [
        ("m", "mute", "Mute Notochord"),
        ("s", "sustain", "Sustain"),
        ("q", "query", "Re-query Notochord"),
        ("r", "reset", "Reset Notochord")]
    
    def compose(self):
        """Create child widgets for the app."""
        yield Header()
        yield self.std_log
        yield NotoLog(id='note')
        yield NotoPrediction(id='prediction')
        yield Mixer()
        yield NotoPresets()
        yield NotoControl()
        yield Footer()

    def on_mount(self) -> None:
        self.query_one(NotoPrediction).tooltip = "displays the next predicted event"
        # print(self.screen)

    def set_preset(self, idx, name):
        node = self.query_one('#'+preset_id(idx))
        node.label = str(idx) if name is None else name
        # node.label = name

    def set_channel(self, chan, cfg):
        # print(f'set_channel {cfg}')
        inst_node = self.query_one('#'+inst_id(chan))
        mode_node = self.query_one('#'+mode_id(chan))
        mute_node = self.query_one('#'+mute_id(chan))

        if cfg is None:
            inst_node.variant = 'default'
            mute_node.variant = 'default'
            mode_node.variant = 'default'
            return
        
        mode = cfg['mode']
        mute = cfg['mute']
        inst = cfg['inst']

        inst_node.label = inst_label(inst)

        if mode=='auto':
            mode_node.label = f"{chan:02d}"
        elif mode=='input':
            mode_node.label = f"-->{chan:02d}"
        elif mode=='follow':
            mode_node.label = f"{cfg['source']:02d}->{chan:02d}"

        if mute:
            mode_node.variant = 'default'
            inst_node.variant = 'default'
            mute_node.label = 'UNMUTE'
            mute_node.variant = 'error'
        else:
            mode_node.variant = 'primary'
            inst_node.variant = 'warning'
            mute_node.label = 'MUTE'
            mute_node.variant = 'default'

gm_names = [
    '_SEQ_\nSTART',
    'GRAND\nPIANO', 'BRGHT\nPIANO', 'EGRND\nPIANO', 'HONKY\n-TONK', 
    'RHODE\nPIANO', 'FM   \nPIANO', 'HRPSI\nCHORD', 'CLAV \n INET',
    'CEL  \n ESTA', 'GLOCN\nSPIEL', 'MUSIC\n BOX ', 'VIBRA\nPHONE', 
    'MAR  \n IMBA', 'XYLO \nPHONE', 'TUBLR\n BELL', 'DULCI\n  MER',
    'DRAWB\nORGAN', 'PERC \nORGAN', 'ROCK \nORGAN', 'CHRCH\nORGAN', 
    'REED \nORGAN', 'ACCOR\n DION', 'HARMO\n NICA', 'BANDO\n NEON',  
    'A-GTR\nNYLON', 'A-GTR\nSTEEL', 'E-GTR\nJAZZ ', 'E-GTR\nCLEAN', 
    'E-GTR\nMUTED', 'E-GTR\nDRIVE', 'E-GTR\nDIST ', 'E-GTR\nHRMNC',  
    'A-GTR\n BASS', 'EBASS\nFINGR', 'EBASS\n PICK', 'EBASS\nFRTLS', 
    'SLAP \nBASS1', 'SLAP \nBASS2', 'SYNTH\nBASS1', 'SYNTH\nBASS2',  
    'STRNG\nVIOLN', 'STRNG\nVIOLA', 'STRNG\nCELLO', 'STRNG\nCBASS', 
    'STRNG\nTREMO', 'STRNG\nPIZZC', 'ORCH \n HARP', 'TIMP \n  ANI',  
    'STRNG\nENSB1', 'STRNG\nENSB2', 'SYNTH\nSTRG1', 'SYNTH\nSTRG2', 
    'CHOIR\n AAH ', 'CHOIR\n OOH ', 'SYNTH\nVOICE', 'ORCH \n  HIT',  
    'TRUM \n  PET', 'TROM \n BONE', 'TUBA \n     ', 'MUTED\nTRMPT', 
    'FRNCH\nHORN ', 'BRASS\nSECTN', 'SYNTH\nBRSS1', 'SYNTH\nBRSS2',  
    'SPRNO\n SAX ', 'ALTO \n  SAX', 'TENOR\n  SAX', 'BARI \n  SAX', 
    'OBOE \n     ', 'ENGLS\nHORN ', 'BASS \n  OON', 'CLARI\n  NET',  
    'PICC \n  OLO', 'FLUTE\n     ', 'RECO \n RDER', ' PAN \nFLUTE',
    'BLOWN\nBOTTL', 'SHAKU\nHACHI', 'WHIS \n  TLE', 'OCA  \n RINA',  
    'LEAD1\nSQUAR', 'LEAD2\n SAW ', 'LEAD3\n TRI ', 'LEAD4\nCHIFF',
    'LEAD5\nCHRNG', 'LEAD6\nVOICE', 'LEAD7\nFIFTH', 'LEAD8\nSYNTH',  
    'PAD 1\nNWAGE', 'PAD 2\n WARM', 'PAD 3\n POLY', 'PAD 4\nCHOIR',
    'PAD 5\nGLASS', 'PAD 6\nMETAL', 'PAD 7\n HALO', 'PAD 8\nSWEEP',
    'FX  1\n RAIN', 'FX  2\nSDTRK', 'FX  3\nCRYST', 'FX  4\nATMOS',
    'FX  5\nBRGHT', 'FX  6\nGOBLN', 'FX  7\nECHOS', 'FX  8\nSCIFI',
    'SITAR\n     ', 'BANJO\n     ', 'SHAM \n ISEN', 'KOTO \n     ',
    'KAL  \n IMBA', 'BAG  \n PIPE', 'FID  \n  DLE', 'SHA  \n  NAI',  
    'TINKL\nBELL ', 'AGO  \n   GÔ', 'STEEL\nDRUM ', 'WOOD \nBLOCK', 
    'TAIKO\nDRUM ', 'MELO \n  TOM', 'SYNTH\nDRUM ', ' REV \nCYMBL',  
    'GTR  \n FRET', 'BRE  \n  ATH', ' SEA \nSHORE', 'BIRD \nTWEET',
    'TELE \nPHONE', 'HELI \nCOPTR', 'APP  \nLAUSE', 'GUN  \n SHOT',  
] + (
    ['STD  \n  KIT']*8 + 
    ['ROOM \n  KIT']*8 + 
    ['ROCK \n  KIT']*8 + 
    ['ELCTR\n  KIT']*8 + 
    ['JAZZ \n  KIT']*8 + 
    ['BRUSH\n  KIT']*8 + 
    ['ORCHS\n  KIT']*8 + 
    ['SFX  \n  KIT']*8 + 
    ['DRUM \n KIT?']*64 
) + ['ANON \n MEL ']*32 + ['ANON \nDRUMS']*32
gm_groups = [
    ('PIANO\n     ', range(1,9)),
    ('CHROM\nPERC ', range(9,17)),
    ('ORGAN\n     ', range(17,25)),
    ('GUI  \n TARS', range(25,33)),
    ('BASS \n GTRS', range(33,41)),
    ('STRIN\n   GS', range(41,49)),
    ('ENSEM\n BLES', range(49,57)),
    ('BRASS\n     ', range(57,65)),
    ('REEDS\n     ', range(65,73)),
    ('PIPES\n     ', range(73,81)),
    ('SYNTH\nLEADS', range(81,89)),
    ('SYNTH\nPADS ', range(89,97)),
    ('SYNTH\nFX   ', range(97,105)),
    ('MISC \n  MEL', range(105,113)),
    ('MISC \n PERC', range(113,121)),
    ('SOUND\nFX   ', range(121,129)),
    ('DRUM \n KITS', [128+i for i in (1,9,17,25,33,41,49,57)]),
    # (' STD \nDRUMS', range(129,137)),
    # ('ROOM \nDRUMS', range(137,145)),
    # ('ROCK \nDRUMS', range(145,153)),
    # ('ELCTR\nDRUMS', range(153,161)),
    # ('JAZZ \nDRUMS', range(161,169)),
    # ('BRUSH\nDRUMS', range(169,177)),
    # ('ORCH \nDRUMS', range(177,185)),
    # (' SFX \nDRUMS', range(185,193)),
    ('ANON\n  MEL', range(257,273)),
    ('ANON\n DRUM', range(289,305)),
]
def inst_label(i):
    if i is None:
        return f"--- \n-----\n-----"
    return f'{i:03d} \n{gm_names[i]}'
### end def TUI components###

if __name__=='__main__':
    run(main)



"""
Notochord + Language Model + Morse Code.

Authors:
  Victor Shepardson
  Intelligent Instruments Lab 2024
"""

import random
# from copy import deepcopy
import warnings

import rich
from rich.text import Text
import torch
from transformers import AutoModelForCausalLM, AutoTokenizer

from notochord import Notochord
from iipyper import MIDI, run, cleanup, repeat, _lock
import time

torch.set_num_threads(1)

def now():
    return time.time()

def main(
        noto_channel=1,
        noto_inst=20, # General MIDI numbered from 1 (see Notochord.feed docstring)
        midi_in=None, # MIDI port for player input
        midi_out=None, # MIDI port for Notochord output
        notochord="notochord-latest.ckpt", # Notochord checkpoint
        # lm="state-spaces/mamba-130m-hf",
        # lm="state-spaces/mamba-790m-hf",
        lm="state-spaces/mamba-1.4b-hf",
        verbose=0,
        prompt='morse code is ',
        dit_dur=0.05,
        send_pc=True,
        lm_temp=1,
        # noto_temp=1.5,
        noto_temp=1,
        noto_space=False, # does notochord contribute to prediction of ' ' char
        buffer_events=8,
        device='cpu',
        ):
    midi = MIDI(midi_in, midi_out)

    print(f'loading notochord model {notochord}')
    noto = Notochord.from_checkpoint(notochord)
    noto.eval()

    # c_dot, c_dash, c_space = '•', '╺━╸', '┃'
    # c_dot, c_dash, c_space = '•', '╺╸', '┃'
    # c_dot, c_dash, c_space = '•', '⁃', '╱'
    c_dot, c_dash, c_space = '.', '╶╴', '╱'

    def do_send_pc(c, i):
        # warn_inst(i)
        # assuming fluidsynth -o synth.midi-bank-select=mma
        if noto.is_drum(i):
            midi.control_change(channel=c-1, control=0, value=1)
        else:
            midi.control_change(channel=c-1, control=32, value=0)
            midi.control_change(channel=c-1, control=0, value=0)
        if noto.is_anon(i):
            program = 0
        else:
            # convert to 0-index
            program = (i-1) % 128
        midi.program_change(channel=c-1, program=program)

    if send_pc:
        do_send_pc(noto_channel, noto_inst)

    print(f'loading language model {lm}')
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        tokenizer = AutoTokenizer.from_pretrained(lm)
        lm = AutoModelForCausalLM.from_pretrained(lm)
        lm.eval()
        lm.to(device)

    tokens = tokenizer.batch_decode(range(len(tokenizer)))
    alphabet = 'abcdefghijklmnopqrstuvwxyzðþöéæ"\'+-=/.,:;!?() '
    alphabet = set(alphabet)
    # indices of within-vocab tokens
    token_idx = [
        i for i,t in enumerate(tokens) if all(c in alphabet for c in t)]
    tokens = [tokens[i] for i in token_idx]
    # token_node_index maps tokens directly to tree nodes for speed
    token_root, token_node_index = TokenNode.make_tree(tokens)

    def set_token_probs(probs):
        for t,p in zip(tokens, probs):
            token_node_index[t].prob = p

    def lm_prompt(prompt):
        """feed a text to the language model and return probabilities for the next token and model state"""
        input_ids = tokenizer(prompt, return_tensors="pt")["input_ids"]
        input_ids = input_ids.to(lm.device)
        with torch.inference_mode():
            state = lm(input_ids, use_cache=True)
            probs = state.logits[0,-1,token_idx].softmax(0).cpu().tolist()
        return probs, state

    def lm_feed_query(token, state):
        """feed a token to the language model and return probabilities for the next token and model state"""
        with torch.inference_mode():
            input_ids = tokenizer(token, return_tensors='pt')['input_ids']
            input_ids = input_ids.to(lm.device)

            # print(input_ids, tokens[input_ids.item()])
            cache = None if state is None else state.cache_params
            # t = now()
            state = lm(input_ids, use_cache=True, cache_params=cache)
            probs = state.logits[0,-1,token_idx].softmax(0).cpu().tolist()
            # print('LM query ms:', int(1000*(now()-t)))

        return probs, state
    
    def set_morse_probs_from_token(morse_tree, token_tree):
        """set the probability of every character in morse to be the sum of probabilities of all currently possible tokens which begin with that character"""
        for node in morse_tree.nodes():
            node.prob = token_tree.children.get(
                node.char, MorseNode()).subtree_prob()
            if verbose > 3:
                print('    setting', node)

    def combine_probs(noto_dt_probs, lm_dt_probs):
        if len(noto_dt_probs)==3:
            noto_dt_probs = [noto_dt_probs[0], sum(noto_dt_probs[1:])]
        return [
            p1**(1/noto_temp) * p2**(1/lm_temp) 
            for p1, p2 in zip(noto_dt_probs, lm_dt_probs)]

    while prompt.startswith(' '):
        print('warning: prompt starts with space')
        prompt = prompt[1:]
    # TODO cache to disk so no delay when prompt is repeated
    # if prompt.endswith(' '):
    #     prompt = prompt[:-1]
    #     initial_space = True
    # else:
    #     initial_space = False
    # TODO get these in absence of prompt?? or they should be None?
    # initial_token_probs, initial_lm_state = lm_prompt(prompt)
    code = MorseNode.make_tree().make_code()
    # initial_morse = ' '.join(
    #     c_space if c==' ' else ''.join(c_dot if e else c_dash for e in code[c])
    #     for c in prompt
    #     )

    prompt_us = [0]
    prompt_token_ends = [False]
    prompt_tokens = iter(
        tokenizer.batch_decode(tokenizer(prompt)["input_ids"]))
    # prompt_u_tokens = [None]
    current_token = next(prompt_tokens)
    idx_in_token = 0
    for c in prompt:
        if c==' ':
            prompt_us[-1] = 7 # make it an interword gap
            # prompt_u_tokens[-1] = (current_token)
        else:
            for e in code[c]:
                prompt_us.append(1 if e else 3) # dit or dah
                prompt_us.append(1) # gap
                # prompt_u_tokens.extend([current_token]*2)
                prompt_token_ends.extend([False]*2)
            prompt_us[-1] = 3 # make it an intercharacter gap

        idx_in_token += 1
        if idx_in_token == len(current_token):
            current_token = next(prompt_tokens, None)
            idx_in_token = 0
            prompt_token_ends[-1] = True

        # print(f'{prompt_tokens=}')
        # print(f'{prompt_us=}')
        # print(f'{prompt_u_tokens=}')
        # print(f'{prompt_token_ends=}')

    class State:
        def __init__(self):
            self.reset()
        def reset(self):
            # self.next_event = None
            # self.next_event_time = None
            self.next_events = []
            self.next_event_times = []
            self.last_event_time = None
            self.noto_held_notes = set()
            self.synth_held_notes = set()

            self.token_probs = None#initial_token_probs
            self.lm_state = None#deepcopy(initial_lm_state)
            self.morse_tree = MorseNode.make_tree()

            self.morse = '' #initial_morse
            self.text = '' #prompt
            self.color_text = Text()
            self.current_token = ''
            self.is_gap = True
            # self.first_event = True
            
            self.token_tree = token_root
            # set_token_probs(self.token_probs)

            # self.token_tree.set_probs(token_idx, self.token_probs)
            # self.token_tree = TokenNode.make_tree(
                # tokens, self.token_probs)
            # self.token_tree = TokenNode.make_tree(
            #     tokens, self.token_probs, alphabet)
            set_morse_probs_from_token(self.morse_tree, self.token_tree)

    S = State()

    def release_all():
        """end any remaining notes"""
        for chan,pitch in S.synth_held_notes:
            midi.note_on(note=pitch, velocity=0, channel=chan-1)

    def reset():
        # TODO expose this
        noto.reset()
        S.reset()
        release_all()

    reset()

    def state_helper(event):
        k = (noto_channel, event['pitch'])
        if event['vel']>0:
            S.noto_held_notes.add(k)
        else:
            try:
                S.noto_held_notes.remove(k)
            except Exception:
                print(f'{k} not in noto_held_notes')
        S.is_gap = not S.is_gap

    def make_query():
        query = {}
        held_pitches = {p for _,p in S.noto_held_notes}
        if S.is_gap:
            # if determining the length of a gap,
            # that means we are predicting the next noteOff
            query['min_vel'] = 1
            query['include_pitch'] = set(range(20,100)) - held_pitches
            # query['next_pitch'] = 40
            # query['include_pitch'] = set(range(20,100)) - held_pitches
        else:
            # if determining the length of the tone,
            # that means we are predicting the next noteOff
            query['next_vel'] = 0
            # print(f'{held_pitches=}')
            if len(held_pitches)>1:
                query['include_pitch'] = held_pitches
            else:
                query['next_pitch'] = list(held_pitches)[0]
        # query['include_inst'] = [1, 8] 
        query['next_inst'] = noto_inst
        return query
    
    def process_event(event, u_possible, noto_dt_probs=None, token_end=None):
        if S.is_gap:
            if u_possible == [0]:
                if verbose > 0:
                    print('initial event')
            elif u_possible == [1]:
                u = 1
                if verbose > 0:
                    print('    character continues')
            else:
                c = S.morse_tree.char
                if verbose > 0:
                    print('  character complete:', c)
                event['text'] = c
                event['morse'] = ' '

                # put the new character into the current token
                S.token_tree = S.token_tree.traverse(c)
                # print(f'{token_tree.token} {"".join(token_tree.children)}')
                # possible end of token + LM query
                if token_end is None:
                    token_end = S.token_tree.sample_terminal(verbose=verbose)
                if token_end:
                    token = S.token_tree.token
                    if verbose > 0:
                        print('token complete:', token)
                    event['token'] = token
                    S.token_probs, S.lm_state = lm_feed_query(
                        token, S.lm_state)
                    S.token_tree = token_root
                    set_token_probs(S.token_probs)

                    if u_possible==[7]:
                        space = True
                    elif u_possible==[3]:
                        space = False
                    else:
                        # now sample nonspace vs space
                        noto_dt_probs = noto_dt_probs[1:]
                        space_prob = S.token_tree.children[' '].subtree_prob()
                        # don't needlessly traverse entire tree to sum probs
                        lm_dt_probs = [
                            sum(S.token_probs)-space_prob, space_prob]

                        if verbose > 1:
                            print(f'      LM prob [nonspace, space] {lm_dt_probs}')

                        if noto_space:
                            probs = combine_probs(noto_dt_probs, lm_dt_probs)
                        else:
                            probs = lm_dt_probs

                        space = random.random() < probs[1] / sum(probs)

                    if space:
                        u = 7
                        if verbose > 0:
                            print('space')
                        S.token_tree = S.token_tree.traverse(' ')
                        event['text'] += ' '
                        event['morse'] += c_space+' '
                    else:
                        u = 3
                        if verbose > 0:
                            print('word continues')

                else:
                    if verbose > 0:
                        print('  token continues')
                    u = 3

                # get next character probs from the token tree
                S.morse_tree = MorseNode.make_tree()
                set_morse_probs_from_token(S.morse_tree, S.token_tree)
                        
        else:
            assert len(u_possible)==1
            u = u_possible[0]
            S.morse_tree = S.morse_tree.traverse(u==1)
            event['morse'] = c_dot if u==1 else c_dash
            if verbose > 0:
                print('    dit' if u==1 else '    dah')

        return u

    def noto_prompt(u, token_end):
        """needs to know the duration of current event, plus whether a token is ending"""
        query = make_query()
        query['next_time'] = u*dit_dur
        # print(f'{u=} {token_end=} {S.is_gap=} {S.morse_tree=} {S.token_tree=}')

        event = noto.query_feed(**query)

        if u!=0:
            process_event(event, [u], token_end=token_end)

        state_helper(event)

        return event

    def noto_query():
        """query notochord + LM as needed for the next MIDI event
        """

        query = make_query()

        if S.is_gap:
            dts = [1,3,7]
        else:
            dts = [1,3]
        dts = torch.tensor([dts])*dit_dur

        with torch.inference_mode():
            event = noto.query(return_params=True, **query)

            dt_breaks = torch.tensor([
                dit_dur*0.5, dit_dur*1.5, dit_dur*4.5])
            cdf = noto.time_dist.cdf(event['time_params'], dt_breaks).squeeze()
            noto_dt_probs = cdf[1:] - cdf[:-1]
            noto_dt_probs = noto_dt_probs.tolist()
            if S.is_gap:
                noto_dt_probs.append(1-cdf[-1].item())

            # time_sample = noto.time_dist(event['time_params'], dts)
            # noto_dt_probs = time_sample['log_prob'].exp().squeeze()
            # noto_dt_probs = [x.item() for x in noto_dt_probs]

        # print(f'{event["vel"]=}, {event["pitch"]=}')

        # get LM dt probs from tree
        if S.is_gap:
            if verbose > 1:
                print(f'      noto prob [1u, 3u, 7u] {noto_dt_probs}')
            # 1: char continues; 3,7: end of char
            # here, the 3,7 probs only need to add up correctly (hence prob/2)
            # we'll check the LM again only if the token ends
            lm_dt_probs = [S.morse_tree.children_prob(), S.morse_tree.prob]
            if verbose > 1:
                print(f'      LM prob [{S.morse_tree.children_chars()}, {S.morse_tree.char}] {lm_dt_probs}')
        else:
            if verbose > 1:
                print(f'      noto prob [1u, 3u] {noto_dt_probs}')
            # 1: dit, 3: dah
            dit = S.morse_tree.traverse(True)
            dah = S.morse_tree.traverse(False)
            lm_dt_probs = [dit.subtree_prob(), dah.subtree_prob()]
            if verbose > 1:
                print(f'      LM prob [dit, dah] [{dit.chars()}, {dah.chars()}] {lm_dt_probs}')

        # print(f'{"gap" if is_gap else "tone"} {lm_dt_probs=}')

        probs = combine_probs(noto_dt_probs, lm_dt_probs)
        if verbose > 1:
            print(f'      combined {probs=}')

        # here we sample 1 vs 3 or 1 vs 3,7
        is_1u = random.random() < probs[0] / (sum(probs) + 1e-10)

        u_possible = [1] if is_1u else ([3,7] if S.is_gap else [3])

        u = process_event(event, u_possible, noto_dt_probs=noto_dt_probs)

        event['time'] = u*dit_dur

        with torch.inference_mode():
            noto.feed(**event)
        
        state_helper(event)

        return event

    @midi.handle(type='program_change')
    def _(msg):
        """
        Program change events set instruments
        """
        nonlocal noto_inst
        if msg.channel == noto_channel:
            noto_inst = msg.program

    @midi.handle(type='control_change', control=0, channel=noto_channel-1)
    def _(msg):
        """
        any CC0 message resets Notochord
        """
        noto.reset()

    @midi.handle(type=('note_on', 'note_off'), channel=noto_channel-1)
    def _(msg):
        """
        MIDI NoteOn events
        """
        if noto is None:
            print('Notochord model not loaded')
            return
    
    # @repeat(1e-3, lock=True)

    color_generator = random.Random()
    def cg():
        return color_generator.randint(150,255)

    @repeat(lock=True)
    def _():
        """Loop, consuming scheduled events"""

        late = now() - S.next_event_times[0] if len(S.next_events) else -1
        if late >= 0:
            if late > 10e-3:
                print(f'{late=}')
            # send MIDI
            event = S.next_events.pop(0)
            pitch = int(event['pitch'])
            vel = int(event['vel'])
            midi.note_on(note=pitch, velocity=vel, channel=noto_channel-1)

            S.morse += event.get('morse', '')
            char = event.get('text', '')
            # S.text += event.get('text', '')

            if 'token' in event:
                tok = event['token']
                color_generator.seed(tok)
                S.color_text.append(event['token'], style=f'rgb({cg()},{cg()},{cg()})')
                S.current_token = ' ' if char.endswith(' ') else ''
            else:
                S.current_token += char

            rich.print(
                Text('\n'+S.morse+'\n')+S.color_text+Text(S.current_token), 
                sep='', end='')

            # print(
                # '\n'+S.morse+'\n'+S.text, 
                # flush=True, end='')

            k = (noto_channel, pitch)
            if vel:
                S.synth_held_notes.add(k)
            else:
                try:
                    S.synth_held_notes.remove(k)
                except Exception:
                    print(f'{k} not in synth_held_notes')

            S.last_event_time = S.next_event_times.pop(0)

        dt = S.next_event_times[0] - now() if len(S.next_event_times) else 10e-3
        return max(0, min(10e-3, dt))

    @repeat(10e-3, lock=False)
    def _():
        """Loop, predicting next events"""          
        if len(S.next_events) < buffer_events:
            final_time = (
                S.next_event_times[-1] if len(S.next_event_times) 
                else S.last_event_time)
            
            # t = now()
            # if final_time is not None:
                # print('buffered time ms:', int(1000*(final_time-t)))
            if len(prompt_us):
                u = prompt_us.pop(0)
                # token = prompt_u_tokens.pop(0)
                token_end = prompt_token_ends.pop(0)
                event = noto_prompt(u, token_end) 
            else:
                event = noto_query()
            # print('query time ms:', int(1000*(now()-t)))

            with _lock:
                S.next_events.append(event)

                if final_time is None:
                    S.next_event_times.append(now() + 2)
                else:
                    S.next_event_times.append(final_time + float(event['time']))

        # print('buffered events:', len(S.next_events))

        # return max(0, min(10e-3, S.next_event_times[0] - now()))
        # return max(0, min(10e-3, S.next_event_time - now()))


    @cleanup
    def _():
        release_all()

class MorseNode:
    def __init__(self, 
            char:str=None, 
            prob:float=0, 
            dit:'MorseNode'=None, 
            dah:'MorseNode'=None
        ):
        self.char = char
        self.prob = prob
        self.dit = dit
        self.dah = dah

    def subtree_prob(self):
        prob = self.prob
        for d in (self.dit, self.dah):
            if d is not None: 
                prob += d.subtree_prob()
        return prob
    
    def children_prob(self):
        return self.subtree_prob() - self.prob
    
    def is_leaf(self):
        return self.dit is None and self.dah is None
    
    def nodes(self):
        nodes = [self]
        for d in (self.dit, self.dah):
            if d is not None:
                nodes.extend(d.nodes())
        return nodes
    
    def chars(self):
        return ''.join(
            MorseNode.char for MorseNode in self.nodes() if MorseNode.char is not None)
    
    def children_chars(self):
        chars = ''
        for d in (self.dit, self.dah):
            if d is not None:
                chars += d.chars()
        return chars
    
    def traverse(self, *seq):
        if not len(seq):
            return self
        b, *seq = seq
        d = self.dit if b else self.dah
        # print(b, seq, d)
        return MorseNode() if d is None else d.traverse(*seq)
        
    def __repr__(self):
        s = f'{self.char} {self.prob}'
        for d in (self.dit, self.dah):
            if d is not None: 
                s += f' {d.char}'
        return s
    
    def make_code(self, path=None):
        """return a mapping from characters to morse"""
        code = {}
        if path is None: path = tuple()
        if self.char is not None:
            code[self.char] = path
        for d,b in ((self.dit, True), (self.dah, False)):
            if d is not None:
                code.update(d.make_code(path+(b,)))
        return code

    @classmethod
    def make_tree(cls):
        return MorseNode(
            dit=MorseNode('e', 12.7,
                dit=MorseNode('i', 7.0,
                    dit=MorseNode('s', 6.3,
                        dit=MorseNode('h', 6.1,
                            dit=MorseNode('5'),
                            dah=MorseNode('4'),
                        ),
                        dah=MorseNode('v', 1.0,
                            # dit=MorseNode('ŝ'),
                            dah=MorseNode('3'),
                        ),
                    ),
                    dah=MorseNode('u', 2.8,
                        dit=MorseNode('f', 2.2,
                            dit=MorseNode('é'),
                            dah=MorseNode('4'),
                        ),
                        dah=MorseNode(#'ü',
                            dit=MorseNode('ð',
                                dit=MorseNode('?')
                            ),
                            dah=MorseNode('2'),
                        ),
                    ),
                ),
                dah=MorseNode('a', 8.2,
                    dit=MorseNode('r', 6.0,
                        dit=MorseNode('l', 4.0,
                            # dit=MorseNode(),
                            dah=MorseNode(#'è',
                                dit=MorseNode('"')
                            ),
                        ),
                        dah=MorseNode('æ',
                            dit=MorseNode('+',
                                dah=MorseNode('.')
                            )
                        ),
                    ),
                    dah=MorseNode('w', 2.4,
                        dit=MorseNode('p', 1.9,
                            dit=MorseNode('þ'),
                            dah=MorseNode(#'à',
                                dit=MorseNode('@'),
                            ),
                        ),
                        dah=MorseNode('j', 0.2,
                            # dit=MorseNode('ĵ'),
                            dah=MorseNode('1',
                                dit=MorseNode("'")
                            ),
                        ),
                    ),
                ),
            ),
            dah=MorseNode('t', 9.1,
                dit=MorseNode('n', 6.7,
                    dit=MorseNode('d', 4.3,
                        dit=MorseNode('b', 1.5,
                            dit=MorseNode('6',
                                dah=MorseNode('-')
                            ),
                            dah=MorseNode('='),
                        ),
                        dah=MorseNode('x', 0.2,
                            dit=MorseNode('/')
                        ),
                    ),
                    dah=MorseNode('k', 0.8,
                        dit=MorseNode('c', 2.8,
                            # dit=MorseNode('ç'),
                            dah=MorseNode(#empty MorseNode
                                dit=MorseNode(';'),
                                dah=MorseNode('!')
                            ),
                        ),
                        dah=MorseNode('y', 2.0,
                            dit=MorseNode('(',
                                dah=MorseNode(')')
                            )
                        ),
                    ),
                ),
                dah=MorseNode('m', 2.4,
                    dit=MorseNode('g', 2.0,
                        dit=MorseNode('z', 0.1,
                            dit=MorseNode('7'),
                            dah=MorseNode(
                                dah=MorseNode(',')
                            ),
                        ),
                        dah=MorseNode('q', 0.1),
                    ),
                    dah=MorseNode('o', 7.5,
                        dit=MorseNode('ö',
                            dit=MorseNode('8',
                                dit=MorseNode(':')
                            ),         
                        ),
                        dah=MorseNode(
                            dit=MorseNode('9'),
                            dah=MorseNode('0')
                        ),
                    ),
                ),
            )
        )

class TokenNode:
    def __init__(self, 
            token:str=None, 
            prob:float=0, 
            children=None,
        ):
        self.token = token
        self.prob = prob
        if children is None:
            children = {}
        self.children = children
            
    def subtree_prob(self):
        prob = self.prob
        for node in self.children.values():
            prob += node.subtree_prob()
        return prob
    
    def children_prob(self):
        return self.subtree_prob() - self.prob
    
    def is_leaf(self):
        return not len(self.children)
    
    def traverse(self, s):
        if not len(s):
            return self
        c, *s = s
        node = self.children.get(c)
        return TokenNode() if node is None else node.traverse(s)
    
    def nodes(self):
        yield self
        for child in self.children.values():
            for node in child.nodes():
                yield node
    
    def sample_terminal(self, verbose=0):
        if verbose>2:
            # print(self)
            print(f'----terminal prob: {self.prob / self.subtree_prob()}')
        if self.prob > 0:
            if random.random() < self.prob / self.subtree_prob():
                return True
        return False
        
    def __repr__(self):
        s = f'{self.token} {self.prob} {"".join(self.children)}'
        return s

    @classmethod
    def make_tree(cls, tokens):
        root = TokenNode()
        index = {}
        for token in tokens:
            tree = root
            for c in token:
                if c not in tree.children:
                    tree.children[c] = TokenNode()
                tree = tree.children[c]
            tree.token = token
            index[token] = tree
        return root, index

if __name__=='__main__':
    run(main)



import math
from typing import List, Tuple, Dict, Union, Any
from numbers import Number
from collections import namedtuple, defaultdict
from pathlib import Path
import hashlib
import json

import mido
from tqdm import tqdm
import appdirs
import numpy as np
import joblib

import torch
from torch import nn
import torch.nn.functional as F
import torch.distributions as D

from .rnn import GenericRNN
from .distributions import CensoredMixtureLogistic, categorical_sample

from .util import arg_to_set, download_url

class Query:
    def __init__(self, modality, then=None, **kw):
        self.modality = modality
        self.then = then
        self.kw = kw
    def __repr__(self):
        return f"{self.modality} \n{self.then}"# {self.kw}"

Range = namedtuple('Range', ('lo', 'hi', 'weight'), defaults=(-torch.inf, torch.inf, 1))
Subset = namedtuple('Subset', ('values', 'weight'), defaults=(None, 1))

def _user_data_dir():
    d = Path(appdirs.user_data_dir('Notochord', 'IIL'))
    d.mkdir(exist_ok=True, parents=True)
    return d

mem = joblib.Memory(_user_data_dir())

class SineEmbedding(nn.Module):
    def __init__(self, n, hidden, w0=1e-3, w1=10, scale='log'):
        """
        Args:
            n (int): number of sinusoids
            hidden (int): embedding size
            w0 (float): minimum wavelength
            w1 (float): maximum wavelength
            scale (str): if 'log', more wavelengths close to w0
        """
        super().__init__()
        if scale=='log':
            w0 = math.log(w0)
            w1 = math.log(w1)
        ws = torch.linspace(w0, w1, n)
        if scale=='log':
            ws = ws.exp()
        self.register_buffer('fs', 2 * math.pi / ws)
        self.proj = nn.Linear(n,hidden)

    def forward(self, x):
        x = x[...,None] * self.fs
        return self.proj(x.sin())

class MixEmbedding(nn.Module):
    def __init__(self, n, domain=(0,1)):
        """
        Args:
            n (int): number of channels
            domain (Tuple[float])
        """
        super().__init__()
        self.domain = domain
        self.lo = nn.Parameter(torch.randn(n))
        self.hi = nn.Parameter(torch.randn(n))
    def forward(self, x):
        """
        Args:
            x: Tensor[...]
        Returns:
            Tensor[...,n]
        """
        x = (x - self.domain[0])/(self.domain[1] - self.domain[0])
        x = x[...,None]
        return self.hi * x + self.lo * (1-x)

class GLU(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        a, b = x.chunk(2, -1)
        return a * b.sigmoid()

class GLUMLP(nn.Module):
    def __init__(self, input, hidden, output, layers, dropout=0, norm=None):
        super().__init__()
        h = input
        def get_dropout():
            if dropout > 0:
                return (nn.Dropout(dropout),)
            return tuple()
        def get_norm():
            if norm=='layer':
                return (nn.LayerNorm(hidden),)
            return tuple()
        self.net = []
        for _ in range(layers):
            self.net.append(nn.Sequential(
                *get_dropout(), nn.Linear(h, hidden*2), GLU(), *get_norm()))
            h = hidden
        self.net.append(nn.Linear(hidden, output))
        self.net = nn.Sequential(*self.net)

        with torch.no_grad():
            self.net[-1].weight.mul_(1e-2)

    def forward(self, x):
        return self.net(x)


class Notochord(nn.Module):
    # note: use named arguments only for benefit of training script
    def __init__(self, 
            emb_size=256, 
            rnn_hidden=2048, rnn_layers=1, kind='gru', 
            mlp_layers=0,
            dropout=0.1, norm=None,
            num_pitches=128, 
            num_instruments=320,
            time_sines=128, vel_sines=128,
            time_bounds=(0,10), time_components=32, time_res=1e-2,
            vel_components=16
            ):
        """
        """
        super().__init__()

        self.note_dim = 4 # instrument, pitch, time, velocity

        self.instrument_start_token = 0
        self.instrument_domain = num_instruments+1

        self.pitch_start_token = num_pitches
        self.pitch_domain = num_pitches+1

        self.max_dt = time_bounds[1]
        self.time_dist = CensoredMixtureLogistic(
            time_components, time_res, 
            sharp_bounds=(1e-4,2e3),
            lo=time_bounds[0], hi=time_bounds[1], init='time')
        self.vel_dist = CensoredMixtureLogistic(
            vel_components, 1.0,
            sharp_bounds=(1e-3,128),
            lo=0, hi=127, init='velocity')
        
        # embeddings for inputs
        self.instrument_emb = nn.Embedding(self.instrument_domain, emb_size)
        self.pitch_emb = nn.Embedding(self.pitch_domain, emb_size)
        self.time_emb = (#torch.jit.script(
            SineEmbedding(
            time_sines, emb_size, 1e-3, 30, scale='log'))
        # self.vel_emb = MixEmbedding(emb_size, (0, 127))
        self.vel_emb = (#torch.jit.script(
            SineEmbedding(
            vel_sines, emb_size, 2, 512, scale='lin'))

        # RNN backbone
        self.rnn = GenericRNN(kind, 
            emb_size, rnn_hidden, 
            num_layers=rnn_layers, batch_first=True, dropout=dropout)

        # learnable initial RNN state
        self.initial_state = nn.ParameterList([
             # layer x batch x hidden
            nn.Parameter(torch.randn(rnn_layers,1,rnn_hidden)*rnn_hidden**-0.5)
            for _ in range(2 if kind=='lstm' else 1)
        ])

        mlp_cls = GLUMLP#lambda *a: torch.jit.script(GLUMLP(*a))
        # projection from RNN state to distribution parameters
        self.h_proj = mlp_cls(
                rnn_hidden, emb_size, emb_size, 
                mlp_layers, dropout, norm)
        self.projections = nn.ModuleList([
            mlp_cls(
                emb_size, emb_size, self.instrument_domain, 
                mlp_layers, dropout, norm),
            mlp_cls(
                emb_size, emb_size, self.pitch_domain, 
                mlp_layers, dropout, norm),
            mlp_cls(
                emb_size, emb_size, self.time_dist.n_params,
                mlp_layers, dropout, norm),
            mlp_cls(
                emb_size, emb_size, self.vel_dist.n_params, 
                mlp_layers, dropout, norm),
        ])

        self.end_proj = nn.Linear(rnn_hidden, 2)

        with torch.no_grad():
            for p in self.projections:
                p.net[-1].weight.mul_(1e-2)
            self.end_proj.weight.mul(1e-2)

        # persistent RNN state for inference
        for n,t in zip(self.cell_state_names(), self.initial_state):
            self.register_buffer(n, t.clone())
        self.step = 0

        # volatile hidden states for caching purposes
        self.h = None
        self.h_query = None      

    def cell_state_names(self):
        return tuple(f'cell_state_{i}' for i in range(len(self.initial_state)))

    @property
    def cell_state(self):
        return tuple(getattr(self, n) for n in self.cell_state_names())

    @property
    def embeddings(self):
        return (
            self.instrument_emb,
            self.pitch_emb,
            self.time_emb,
            self.vel_emb
        )
        
    def forward(self, instruments, pitches, times, velocities, ends,
            validation=False, ar_mask=None):
        """
        teacher-forced probabilistic loss and diagnostics for training.

        Args:
            instruments: LongTensor[batch, time]
            pitches: LongTensor[batch, time]
            times: FloatTensor[batch, time]
            velocities: FloatTensor[batch, time]
            ends: LongTensor[batch, time]
            validation: bool (computes some extra diagnostics)
            ar_mask: Optional[Tensor[note_dim x note_dim]] if None, generate random
                masks for training
        """
        batch_size, batch_len = pitches.shape

        self.checkpoint_path = None

        # embed data to input vectors
        inst_emb = self.instrument_emb(instruments) # batch, time, emb_size
        pitch_emb = self.pitch_emb(pitches) # batch, time, emb_size
        time_emb = self.time_emb(times) # batch, time, emb_size
        vel_emb = self.vel_emb(velocities) # batch, time, emb_size

        embs = (inst_emb, pitch_emb, time_emb, vel_emb)

        # feed to RNN backbone
        x = sum(embs)
        ## broadcast initial state to batch size
        initial_state = tuple(
            t.expand(self.rnn.num_layers, x.shape[0], -1).contiguous() # 1 x batch x hidden
            for t in self.initial_state)
        h, _ = self.rnn(x, initial_state) #batch, time, hidden_size

        # fit all event factorizations 
        # e.g. inst->pitch->time->vel vs vel->time->inst->pitch
        trim_h = h[:,:-1]
        # always include hidden state, never include same modality,
        # other dependencies are random per time and position
        n = self.note_dim
        if ar_mask is None:
            # random binary mask
            ar_mask = torch.randint(2, (*trim_h.shape[:2],n,n), dtype=torch.bool, device=h.device)
            # zero diagonal
            ar_mask &= ~torch.eye(n,n, dtype=torch.bool, device=h.device)
        # include hidden state
        ar_mask = torch.cat((ar_mask.new_ones(*ar_mask.shape[:-2],1,n), ar_mask), -2).float()

        to_mask = torch.stack((
            self.h_proj(trim_h),
            *(emb[:,1:] for emb in embs)
        ), -1)
        # TODO: try without this tanh?
        mode_hs = (to_mask @ ar_mask).tanh().unbind(-1)
        
        # final projections to raw distribution parameters
        inst_params, pitch_params, time_params, vel_params = [
            proj(h) for proj,h in zip(self.projections, mode_hs)]

        # get likelihood of data for each modality
        inst_logits = F.log_softmax(inst_params, -1)
        inst_targets = instruments[:,1:,None] #batch, time, 1
        inst_log_probs = inst_logits.gather(-1, inst_targets)[...,0]

        pitch_logits = F.log_softmax(pitch_params, -1)
        pitch_targets = pitches[:,1:,None] #batch, time, 1
        pitch_log_probs = pitch_logits.gather(-1, pitch_targets)[...,0]

        time_targets = times[:,1:] # batch, time
        time_result = self.time_dist(time_params, time_targets)
        time_log_probs = time_result.pop('log_prob')

        vel_targets = velocities[:,1:] # batch, time
        vel_result = self.vel_dist(vel_params, vel_targets)
        vel_log_probs = vel_result.pop('log_prob')

        # end prediction
        # skip the first position for convenience 
        # (so masking is the same for end as for note parts)
        end_params = self.end_proj(h[:,1:])
        end_logits = F.log_softmax(end_params, -1)
        end_log_probs = end_logits.gather(-1, ends[:,1:,None])[...,0]

        r = {
            'end_log_probs': end_log_probs,
            'instrument_log_probs': inst_log_probs,
            'pitch_log_probs': pitch_log_probs,
            'time_log_probs': time_log_probs,
            'velocity_log_probs': vel_log_probs,
            **{'time_'+k:v for k,v in time_result.items()},
            **{'velocity_'+k:v for k,v in vel_result.items()}
        }
        # this just computes some extra diagnostics which are inconvenient to do in the
        # training script. should be turned off during training for performance.
        if validation:
            with torch.no_grad():
                r['time_acc_30ms'] = (
                    self.time_dist.cdf(time_params, time_targets + 0.03)
                    - torch.where(time_targets - 0.03 >= 0,
                        self.time_dist.cdf(time_params, time_targets - 0.03),
                        time_targets.new_zeros([]))
                )
        return r


    # 0 - start token
    # 1-128 - melodic
    # 129-256 - drums
    # 257-288 - anon melodic
    # 289-320 - anon drums
    def is_drum(self, inst):
        # TODO: add a constructor argument to specify which are drums
        # hardcoded for now
        return inst > 128 and inst < 257 or inst > 288
    def is_anon(self, inst):
        return inst > 256
    def first_anon_like(self, inst):
        # TODO: add a constructor argument to specify how many anon
        # hardcoded for now
        return 289 if self.is_drum(inst) else 257
    
    def feed(self, inst:int, pitch:int, time:Number, vel:Number, **kw):
        """consume an event and advance hidden state
        
        Args:
            inst: int. instrument of current note.
                0 is start token
                1-128 are General MIDI instruments
                129-256 are drumkits (MIDI 1-128 on channel 13)
                257-288 are 'anonymous' melodic instruments
                289-320 are 'anonymous' drumkits
            pitch: int. MIDI pitch of current note.
                0-127 are MIDI pitches / drums
                128 is start token
            time: float. elapsed time in seconds since previous event.
            vel: float. (possibly dequantized) MIDI velocity from 0-127 inclusive.
                0 indicates a note-off event
            **kw: ignored (allows doing e.g. noto.feed(**noto.query(...)))
        """
        # print(f'FEED from {threading.get_ident()}') 

        with torch.inference_mode():
            inst = torch.LongTensor([[inst]]) # 1x1 (batch, time)
            pitch = torch.LongTensor([[pitch]]) # 1x1 (batch, time)
            time = torch.FloatTensor([[time]]) # 1x1 (batch, time)
            vel = torch.FloatTensor([[vel]]) # 1x1 (batch, time)

            embs = [
                self.instrument_emb(inst), # 1, 1, emb_size
                self.pitch_emb(pitch), # 1, 1, emb_size
                self.time_emb(time),# 1, 1, emb_size
                self.vel_emb(vel)# 1, 1, emb_size
            ]
            x = sum(embs)

            self.h, new_state = self.rnn(x, self.cell_state)
            for t,new_t in zip(self.cell_state, new_state):
                t[:] = new_t

            self.h_query = None
            
            self.step += 1


    # TODO: add end prediction to deep_query
    def deep_query(self, query, predict_end=True):
        """flexible querying with nested Query objects.
        see query_vtip for an example.

        Args:
            query: Query object
        """
        if self.h_query is None:
            with torch.inference_mode():
                self.h_query = self.h_proj(self.h)
        event = self._deep_query(
            query, hidden=self.h_query[:,0], event={})
        
        if predict_end:
            # print('END')
            # print(f'{self.h}')
            with torch.inference_mode():
                end_params = self.end_proj(self.h)
                event['end'] = end_params.softmax(-1)[...,1].item()
                # event['end'] = D.Categorical(logits=end_params).sample().item()
        else:
            event['end'] = 0#torch.zeros(self.h.shape[:-1])

        return event
    
    def _deep_query(self, query, hidden, event):
        m = query.modality
        sq = query.then
        try:
            idx = ('inst','pitch','time','vel').index(m)
        except ValueError:
            raise ValueError(f'unknown modality "{m}"')
        
        project = self.projections[idx]
        embed = self.embeddings[idx]

        if m=='time':
            dist = self.time_dist
            sample = dist.sample
        elif m=='vel':
            dist = self.vel_dist
            sample = dist.sample
        else:
            sample = categorical_sample

        # Query.then can be:

        # None (no further sub-queries)
        # or str (no further sub-queries, set the 'path' property of the event)

        # Query describing what to do next

        # pair of Number, Query
        # setting a deterministic value for the current query,
        # and the query to execute next

        # list of Range, Query pairs
        # or list of Subset, Query pairs
        #    describing what to do next given where the sampled value falls
        #    Range and Subset can also carry weights


        with torch.inference_mode():
            #
            if sq is None or isinstance(sq, str) or isinstance(sq, Query):
                # print(f'{m=} (simple)')
                # this is the final query (sq is None or a path tag)
                # or there are no cases, just proceed to next subquery
                action = sq
                params = project(hidden.tanh())
                result = sample(params, **query.kw)

            elif len(sq)==0:
                raise ValueError(f"subquery has no cases: {sq}")
            
            # deterministic single value cases
            # elif len(sq)==1 and len(sq[0])==3 and isinstance(sq[0][0], Number):
            #     # singleton list case
            #     key, action = sq[0]
            #     result = key
            elif len(sq)==2 and isinstance(sq[0], Number):
                # print(f'{m=} ({key=})')
                # bare pair case
                key, action = sq
                result = key

            elif m in ('inst', 'pitch'):
                # print(f'{m=}')

                # weighted subsets case
                try:
                    assert all(isinstance(s, Subset) for s,_ in sq)
                except Exception:
                    raise ValueError(f"""
                    subqueries should all be Subset, Query tuples here: {sq}
                    """)

                # if (~hidden.isfinite()).any():
                #     raise Exception('hidden '+str(hidden))
                
                params = project(hidden.tanh())

                # if (~params.isfinite()).any():
                    # raise Exception('params '+str(params))

                 # sample subset
                if len(sq) > 1:
                    all_probs = params.softmax(-1)
                    probs = [
                        all_probs[...,s.values].sum(-1) * s.weight
                        for s,_ in sq]
                    
                    # if (~torch.tensor(probs).log().isfinite()).any():
                        # raise Exception('subset '+str(probs))
                    
                    # print(f'deep_query {m} {sq=} {probs=}')
                    idx = categorical_sample(torch.tensor(probs).log())
                else:
                    idx = 0
                s,action = sq[idx]
                # sample from range
                result = sample(params, whitelist=s.values, **query.kw)

            elif m in ('time', 'vel'):
                # print(f'{m=}')
                # weighted ranges case
                assert all(isinstance(r, Range) for r,_ in sq)

                params = project(hidden.tanh())
                # sample range
                if len(sq) > 1:
                    probs = [
                        (dist.cdf(params, r.hi) - dist.cdf(params, r.lo)
                        ) * r.weight
                        for r,_ in sq]
                    # print(f'deep_query {m} {probs=}')
                    idx = categorical_sample(torch.tensor(probs).log())
                else:
                    idx = 0
                r,action = sq[idx]
                # sample from range
                result = sample(params, truncate=(r.lo, r.hi), **query.kw)

            if not result.isfinite().all():
                print('WARNING: nonfinite value {result=} {m=}')
                result.nan_to_num_(0)

            try:
                event[m] = result.item()
            except Exception:
                event[m] = result

            # print(f'{result=}')
            # embed, add to hidden, recurse into subquery
            if isinstance(action, Query):
                emb = embed(result)
                hidden = hidden + emb
                if (~hidden.isfinite()).any():
                    raise Exception(f'{m=} {result=} {emb=}')
                return self._deep_query(action, hidden, event)
            else:
                event['path'] = action
                return event
            
    def query_tipv_onsets(self,
        min_time=None, max_time=None, 
        include_inst=None,
        include_pitch=None,
        truncate_quantile_time=None,
        truncate_quantile_pitch=None,
        rhythm_temp=None, timing_temp=None,
        min_vel=None, max_vel=None
        ):
        """
        for onset-only_models
        """
        q = Query(
            'time',
            truncate=(min_time or -torch.inf, max_time or torch.inf), 
            truncate_quantile=truncate_quantile_time,
            weight_top_p=rhythm_temp, component_temp=timing_temp,
            then=Query(
                'inst',
                whitelist=include_inst,
                then=Query(
                    'pitch',
                    whitelist=include_pitch,
                    truncate_quantile=truncate_quantile_pitch,
                    then=Query(
                        'vel',
                        truncate=(min_vel or 0.5, max_vel or torch.inf),
                    )
                )
            )
        )
        return self.deep_query(q)
    
    def query_itpv_onsets(self,
        min_time=None, max_time=None, 
        include_inst=None,
        include_pitch=None,
        truncate_quantile_time=None,
        truncate_quantile_pitch=None,
        rhythm_temp=None, timing_temp=None,
        min_vel=None, max_vel=None
        ):
        """
        for onset-only_models
        """
        q = Query(
            'inst',
            whitelist=include_inst,
            then=Query(
                'time',
                truncate=(min_time or -torch.inf, max_time or torch.inf), 
                truncate_quantile=truncate_quantile_time,
                weight_top_p=rhythm_temp, component_temp=timing_temp,
                then=Query(
                    'pitch',
                    whitelist=include_pitch,
                    truncate_quantile=truncate_quantile_pitch,
                    then=Query(
                        'vel',
                        truncate=(min_vel or 0.5, max_vel or torch.inf),
                    )
                )
            )
        )
        return self.deep_query(q)


    def query_vtip(self,
            note_on_map:Dict[int,List[int]], 
            note_off_map:Dict[int,List[int]], 
            min_time=None, max_time=None, 
            rhythm_temp=None, timing_temp=None,
            truncate_quantile_time=None,
            truncate_quantile_pitch=None,
            steer_density=None, # truncate_quantile_type ? 
            inst_weights=None,
            no_steer=None, # TODO
            ):
        """ Query in velocity-time-instrument-pitch order,
            efficiently truncating the joint distribution to just allowable
            (velocity>0)/instrument/pitch triples.

            Args:
                note_on_map: possible note-ons as {instrument: [pitch]} 
                note_off_map: possible note-offs as {instrument: [pitch]} 
        """
        no_on = all(len(ps)==0 for ps in note_on_map.values())
        no_off = all(len(ps)==0 for ps in note_off_map.values())
        if no_on and no_off:
            raise ValueError(f"""
                no possible notes {note_on_map=} {note_off_map=}""")
        
        def get_subquery(ipm, path, weights=None):
            return Query(
                'time',
                truncate=(min_time or -torch.inf, max_time or torch.inf),
                truncate_quantile=truncate_quantile_time,
                weight_top_p=rhythm_temp, component_temp=timing_temp,
                then=Query(
                    'inst', [(
                        Subset([i], 1 if weights is None else weights[i]), 
                        Query('pitch', path, 
                            whitelist=list(ps), 
                            truncate_quantile=None if self.is_drum(i) else truncate_quantile_pitch)
                    ) for i,ps in ipm.items() if len(ps)]
                ),
            )
        w = 1 if steer_density is None else 2**(steer_density*2-1)
        
        w_on = 0 if no_on else w
        w_off = 0 if no_off else 1/w
        
        return self.deep_query(Query('vel', [
            (Range(-torch.inf,0.5,w_off), get_subquery(note_off_map, 'note off')),
            (Range(0.5,torch.inf,w_on), get_subquery(note_on_map, 'note on', inst_weights))
        ]))
    
    def query_vipt(self,
        note_on_map, note_off_map,
        min_time=None, max_time=None, 
        rhythm_temp=None, timing_temp=None,
        truncate_quantile_time=None,
        truncate_quantile_pitch=None,
        steer_density=None,
        inst_weights=None,
        no_steer=None,
        ):
        """
        Query in velocity-instrument-pitch-time order,
            efficiently truncating the joint distribution to just allowable
            (velocity>0)/instrument/pitch triples.

            Args:
                note_on_map: possible note-ons as {instrument: [pitch]} 
                note_off_map: possible note-offs as {instrument: [pitch]} 
        """
        no_on = all(len(ps)==0 for ps in note_on_map.values())
        no_off = all(len(ps)==0 for ps in note_off_map.values())
        if no_on and no_off:
            raise ValueError(f"""
                no possible notes {note_on_map=} {note_off_map=}""")
        
        def get_subquery(note_map, path, weights=None):
            return Query(
                'inst', 
                then=[(Subset([i], 1 if weights is None else weights[i]), Query(
                    'pitch', 
                    whitelist=list(ps), 
                    truncate_quantile=None if self.is_drum(i) else truncate_quantile_pitch,
                    then=Query(
                        'time', path,         
                        truncate=(min_time or -torch.inf, max_time or torch.inf), 
                        truncate_quantile=None if (no_steer is not None and i in no_steer) else truncate_quantile_time,
                        weight_top_p=rhythm_temp, component_temp=timing_temp
                    )
                )) for i,ps in note_map.items() if len(ps)]
            )
        
        w = 1 if steer_density is None else 2**(steer_density*2-1)
        
        w_on = 0 if no_on else w
        w_off = 0 if no_off else 1/w
        
        return self.deep_query(Query('vel', [
            (Range(-torch.inf,0.5,w_off), get_subquery(note_off_map, 'note off')),
            (Range(0.5,torch.inf,w_on), get_subquery(note_on_map, 'note on', inst_weights))
        ]))
    
    # def query_ipvt(self,
    #     note_map, 
    #     min_time=-torch.inf, max_time=torch.inf, 
    #     min_vel=-torch.inf, max_vel=torch.inf,
    #     truncate_quantile_time=None,
    #     truncate_quantile_pitch=None,
    #     ):
    #     """
    #     """
      
    #     return self.deep_query(Query(
    #         'inst', then=[(
    #             Subset([i]), Query(
    #                 'pitch', 
    #                 whitelist=list(ps), 
    #                 truncate_quantile=truncate_quantile_pitch,
    #                 then=Query(
    #                     'vel',
    #                     truncate=(min_vel or -torch.inf, max_vel or torch.inf),
    #                     then=Query(
    #                         'time',         
    #                         truncate=(min_time or -torch.inf, max_time or torch.inf), truncate_quantile=truncate_quantile_time
    #                     )
    #                 )
    #             )
    #         ) for i,ps in note_map.items() if len(ps)]
    #     ))

    # TODO: remove pitch_topk and sweep_time?
    # TODO: rewrite this to build queries and dispatch to deep_query
    def query(self,
            next_inst:int=None, next_pitch:int=None, 
            next_time:float=None, next_vel:int=None,

            allow_end:bool=False,

            include_inst:List[int]=None, exclude_inst:List[int]=None,
            allow_anon:bool=True, 
            instrument_temp:float=None, 

            include_pitch:List[int]=None, exclude_pitch:List[int]=None,
            include_drum:List[int]=None,
            truncate_quantile_pitch:Tuple[float,float]=None,
            pitch_temp:float=None, 
            index_pitch:int=None,

            min_time:float=None, max_time:float=None,
            truncate_quantile_time:Tuple[float, float]=None,
            rhythm_temp:float=None, timing_temp:float=None,

            min_vel:int=None, max_vel:int=None,
            velocity_temp:float=None,

            pitch_topk:int=None, sweep_time:bool=False, 

            handle:str=None, return_params:bool=False
            ) -> dict:
        """
        return a prediction for the next note.

        various constraints on the the next note can be requested.

        Args:
            # hard constraints

            next_inst: fix a particular instrument for the predicted note.
                sampled values will always condition on fixed values, so passing
                `next_inst=1`, for example, will make the event appropriate
                for the piano (instrument 1) to play.
            next_pitch: fix a particular MIDI number for the predicted note.
                sampled values will always condition on fixed values, so passing
                `next_pitch=60`, for example, will make the event appropriate
                for a middle C.
            next_time: fix a particular delta time for the predicted note.
                sampled values will always condition on fixed values, so passing
                `next_time=0`, for example, will make the event appropriate
                for one which is concurrent with the previous event.
            next_vel: fix a particular velocity for the predicted note.
                sampled values will always condition on fixed values, so passing
                `next_inst=0`, for example, will make the event appropriate
                for a noteOff (i.e., something which is currently playing).
                
            # partial constraints

            include_inst: instrument id(s) to include in sampling.
                (if not None, all others will be excluded)
            exclude_inst: instrument id(s) to exclude from sampling.
            allow_anon: bool. if False, zero probability of anon instruments

            include_pitch: pitch(es) to include in sampling.
                (if not None, all others will be excluded)
            exclude_pitch: pitch(es) to exclude from sampling.
            include_drum: like `include_pitch`, but only in effect when 
                instrument is a drumkit

            min_time: if not None, truncate the time distribution below
            max_time: if not None, truncate the time distribution above

            min_vel: if not None, truncate the velocity distribution below
                e.g., `min_vel=1` prevents NoteOff events
            max_vel: if not None, truncate the velocity distribution above

            allow_end: if False, zero probability of sampling the end marker

            # sampling strategies
            
            instrument_temp: if not None, apply top_p sampling to instrument. 0 is
                deterministic, 1 is 'natural' according to the model

            pitch_temp: if not None, apply top_p sampling to pitch. 0 is
                deterministic, 1 is 'natural' according to the model
            truncate_quantile_pitch: applied after include_pitch, exclude_pitch
                truncate the remaining pitch distribution by quantile.
                e.g. truncate_quantile_pitch=(0.25, 0.75)
                excludes the lowest- and highest- pitch 25% of probability mass
            index_pitch: if not None, deterministically take the
                nth most likely pitch instead of sampling.

            timing_temp: if not None, apply temperature sampling to the time
                component. this affects fine timing; 0 is deterministic and 
                precise, 1 is 'natural' according to the model.
            rhythm_temp: if not None, apply top_p sampling to the weighting
                of mixture components. this affects coarse rhythmic patterns;
                0 is deterministic, 1 is 'natural' according to the model
            truncate_quantile_time: applied after min_time, max_time
                truncate the remaining delta time distribution by quantile.
                e.g. truncate_quantile_time=(0.25, 0.75)
                excludes the soonest- and furthest- 25% of probability mass

            velocity_temp: if not None, apply temperature sampling to the velocity
                component.

            # multiple predictions

            pitch_topk: Optional[int]. if not None, instead of sampling pitch, 
                stack the top k most likely pitches along the batch dimension
            sweep_time: if True, instead of sampling time, choose a diverse set
            of times and stack along the batch dimension

            # other

            handle: metadata to be included in the returned dict, if not None
            return_params: if True, return tensors of distribution parameters
                under the keys `inst_params`, `pitch_params`, `time_params`,
                and `vel_params`.

        Returns:
            'inst': int. id of predicted instrument.
                1-128 are General MIDI standard melodic instruments
                129-256 are drumkits for MIDI programs 1-128
                257-288 are 'anonymous' melodic instruments
                289-320 are 'anonymous' drumkits
            'pitch': int. predicted MIDI number of next note, 0-128.
            'time': float. predicted time to next note in seconds.
            'vel': float. unquantized predicted velocity of next note.
                0-127; hard 0 indicates a note-off event.
            'end': int. value of 1 indicates the *current* event (the one 
                passed as arguments to `predict`) was the last event, and the
                predicted event should *not* be played. if `allow end` is false, 
                this will always be 0.
            'step': int. number of steps since calling `reset`.
            '*_params': tensor. distribution parameters for visualization
                and debugging purposes. present if `return_params` is True.

        NOTE: `instrument`, `pitch`, `time`, `velocity` may return lists,
            when using `sweep_time` or `pitch_topk`. that part of the API 
            is very experimental and likely to break.
        """
         # validate options:
        if (index_pitch is not None) and (pitch_temp is not None):
            print("warning: `index pitch` overrides `pitch_temp`")

        inst_intervention = any(p is not None for p in (
            instrument_temp, include_inst, exclude_inst))

        pitch_intervention = (pitch_topk or any(p is not None for p in (
            pitch_temp, include_pitch, exclude_pitch, include_drum)))

        time_intervention = any(p is not None for p in (
            min_time, max_time, rhythm_temp, timing_temp))

        vel_intervention = any(p is not None for p in (
            min_vel, max_vel, velocity_temp))

        exclude_inst = arg_to_set(exclude_inst)
        if not allow_anon:
            exclude_inst |= set(range(257, 321))
        constrain_inst = list((
            set(range(self.instrument_domain)) - {self.instrument_start_token}
            if include_inst is None 
            else arg_to_set(include_inst)
        ) - exclude_inst)
        if len(constrain_inst)==0:
            raise ValueError("""
            every instrument has been excluded. check values of 
            `include_inst` and `exclude_inst`
            """)
        # elif len(constrain_inst)==1:
        #     print("""
        #     warning: recommended to use `next_inst`, not 
        #     `include_inst` to allow only one specific instrument
        #     """)
        
        constrain_pitch = list((
            set(range(self.pitch_domain)) - {self.pitch_start_token}
            if include_pitch is None 
            else arg_to_set(include_pitch)
        ) - arg_to_set(exclude_pitch))
        if len(constrain_pitch)==0:
            raise ValueError("""
            every pitch has been excluded. check values of 
            `include_pitch` and `exclude_pitch`
            """)
        elif len(constrain_pitch)==1:
            print("""
            warning: recommended to use `next_pitch`, not 
            `include_pitch` to allow only one specific pitch
            """)

        # TODO: this got really complicated to support include_drum...
        # really want to edit the whole joint distribution of pitch,inst in 
        # cases where certain pitches or drums need to be excluded...
        # would that be practical? if there are ~40000 inst x pitch combos?
        # would need to run the instrument head for a whole batch of all
        # allowable pitches or vice-versa...
        def sample_instrument(x):
            # if include_drum is supplied, make sure to exclude drum instruments
            # when no pitch is in the allowed drums
            if include_drum is not None:
                pit = predicted_by_name('pitch')
                pits = [pit] if pit is not None else constrain_pitch
                if pits is not None and all(pit not in include_drum for pit in pits):
                    nonlocal constrain_inst
                    if constrain_inst is None:
                        constrain_inst = range(1,self.instrument_domain)
                    constrain_inst = [
                        i for i in constrain_inst if not self.is_drum(i)]

            # if constrain_inst is not None:
            #     preserve_x = x[...,constrain_inst]
            #     x = torch.full_like(x, -torch.inf)
            #     x[...,constrain_inst] = preserve_x
            # probs = x.softmax(-1)
            # if instrument_temp is not None:
            #     probs = reweight_top_p(probs, instrument_temp)
            # return D.Categorical(probs).sample()

            return categorical_sample(x, 
                whitelist=constrain_inst,
                top_p=instrument_temp)

        def sample_pitch(x):
            # conditional constraint
            if include_drum is not None:
                # if this event is / must be a drum,
                # use include_drum instead of constrain_inst
                inst = predicted_by_name('instrument')
                insts = [inst] if inst is not None else constrain_inst
                if insts is not None and all(self.is_drum(i) for i in insts):
                    nonlocal constrain_pitch
                    constrain_pitch = include_drum

            if pitch_topk is not None:
                raise NotImplementedError

            return categorical_sample(x,
                whitelist=constrain_pitch, 
                index=index_pitch,
                top_p=pitch_temp,
                truncate_quantile=truncate_quantile_pitch
                )
            # if constrain_pitch is not None:
            #     preserve_x = x[...,constrain_pitch]
            #     x = torch.full_like(x, -torch.inf)
            #     x[...,constrain_pitch] = preserve_x
            # # x is modified logits

            # if index_pitch is not None:
            #     return x.argsort(-1, True)[...,index_pitch]
            # elif pitch_topk is not None:
            #     return x.argsort(-1, True)[...,:pitch_topk].transpose(0,-1)
            
            # probs = x.softmax(-1)
            # if pitch_temp is not None:
            #     probs = reweight_top_p(probs, pitch_temp)

            # if steer_pitch is not None:
            #     return steer_categorical(probs, steer_pitch)
            # else:
            #     return D.Categorical(probs).sample()

        def sample_time(x):
            # TODO: respect trunc_time when sweep_time is True
            if sweep_time:
                if min_time is not None or max_time is not None:
                    raise NotImplementedError("""
                    min_time/max_time with sweep_time needs implementation
                    """)
                assert x.shape[0]==1, "batch size should be 1 here"
                log_pi, loc, s = self.time_dist.get_params(x)
                idx = log_pi.squeeze().argsort()[:9]
                loc = loc.squeeze()[idx].sort().values[...,None] 
                # multiple times in batch dim
                # print(loc.shape)
                return loc
            
            trunc = (
                -torch.inf if min_time is None else min_time,
                torch.inf if max_time is None else max_time)

            return self.time_dist.sample(x, 
                truncate=trunc,
                component_temp=timing_temp, 
                weight_top_p=rhythm_temp,
                truncate_quantile=truncate_quantile_time
                )

        def sample_velocity(x):
            trunc = (
                -torch.inf if min_vel is None else min_vel,
                torch.inf if max_vel is None else max_vel)
            return self.vel_dist.sample(
                x, component_temp=velocity_temp, truncate=trunc,
                # truncate_quantile=truncate_quantile_vel
                )

        with torch.inference_mode():
            if self.h_query is None:
                self.h_query = self.h_proj(self.h)

            modalities = list(zip(
                self.projections,
                (sample_instrument, sample_pitch, sample_time, sample_velocity),
                self.embeddings,
                ))

            context = [self.h_query] # embedded outputs for autoregressive prediction
            predicted = [] # raw outputs
            params = [] # distribution parameters for visualization

            fix = [
                None if item is None else torch.tensor([[item]], dtype=dtype)
                for item, dtype in zip(
                    [next_inst, next_pitch, next_time, next_vel],
                    [torch.long, torch.long, torch.float, torch.float])]

            # if any modalities are determined, embed them
            # sort constrained modalities before unconstrained
            # TODO: option to skip modalities
            det_idx, cons_idx, uncons_idx = [], [], []
            for i,(item, embed) in enumerate(zip(fix, self.embeddings)):
                if item is None:
                    if (
                        i==0 and inst_intervention or
                        i==1 and pitch_intervention or
                        i==2 and time_intervention or
                        i==3 and vel_intervention):
                        cons_idx.append(i)
                    else:
                        uncons_idx.append(i)
                else:
                    det_idx.append(i)
                    context.append(embed(item))
                    predicted.append(item)
                    params.append(None)
            undet_idx = cons_idx + uncons_idx
            perm = det_idx + undet_idx # permutation from the canonical order
            iperm = np.argsort(perm) # inverse permutation back to canonical order

            mode_names = ['instrument', 'pitch', 'time', 'velocity']
            name_to_idx = {k:v for k,v in zip(mode_names, iperm)}
            def predicted_by_name(name):
                idx = name_to_idx[name]
                if len(predicted) > idx:
                    return predicted[idx]
                return None
            # print('sampling order:', [mode_names[i] for i in perm])

            # for each undetermined modality, 
            # sample a new value conditioned on already determined ones
            
            running_ctx = sum(context)
            # print(running_ctx)
            # perm_h_tgt = [h_tgt[i] for i in perm]
            while len(undet_idx):
                # print(running_ctx.norm())
                i = undet_idx.pop(0) # index of modality to determine
                # j = len(det_idx) # number already determined
                project, sample, embed = modalities[i]
                # determine value for the next modality
                hidden = running_ctx.tanh()
                params.append(project(hidden))
                pred = sample(params[-1])
                predicted.append(pred)
                # prepare for next iteration
                if len(undet_idx):
                    # context.append(embed(pred))
                    running_ctx += embed(pred)
                det_idx.append(i)

            pred_inst = predicted_by_name('instrument')
            pred_pitch = predicted_by_name('pitch')
            pred_time = predicted_by_name('time')
            pred_vel = predicted_by_name('velocity')

            if allow_end:
                end_params = self.end_proj(self.h)
                # print(end_params)
                end = D.Categorical(logits=end_params).sample()
            else:
                end = torch.zeros(self.h.shape[:-1])

            if sweep_time or pitch_topk:
                # return lists of predictions
                pred_inst = [x.item() for x in pred_inst]
                pred_pitch = [x.item() for x in pred_pitch]
                pred_time = [x.item() for x in pred_time]
                pred_vel = [x.item() for x in pred_vel]
                end = [x.item() for x in end]
                # print(pred_time, pred_pitch, pred_vel)
            else:
                # return single predictions
                pred_inst = pred_inst.item()
                pred_pitch = pred_pitch.item()
                pred_time = pred_time.item()
                pred_vel = pred_vel.item()
                end = end.item()

            r = {
                'inst': pred_inst,
                'pitch': pred_pitch, 
                'time': pred_time,
                'vel': pred_vel,

                'end': end,
                'step': self.step,
            }

            if handle is not None:
                r['handle'] = handle

            if return_params:
                r |= {
                    'inst_params': params[iperm[0]],
                    'pitch_params': params[iperm[1]],
                    'time_params': params[iperm[2]],
                    'vel_params': params[iperm[3]]
                }

            return r

    def predict(self, inst, pitch, time, vel, **kw):
        """
        DEPRECATED: alias for feed_query
        """
        self.feed(inst, pitch, time, vel)
        return self.query(**kw)

    def feed_query(self, inst:int, pitch:int, time:Number, vel:Number, 
 **kw):
        """
        feed an event to the model, 
        then query for the next predicted event and return it.
        """
        self.feed(inst, pitch, time, vel)
        return self.query(**kw)

    def query_feed(self, *a, **kw):
        """
        query for the next predicted event and immediately feed it to the model,
        also returning the predicted event.
        """
        r = self.query(*a, **kw)
        self.feed(r['inst'], r['pitch'], r['time'], r['vel'])
        return r

    def feed_query_feed(self, 
            inst:int, pitch:int, time:Number, vel:Number, 
            **kw):
        """
        given an event, return the next predicted event, 
        feeding both to the model.
        """
        self.feed(inst, pitch, time, vel)
        return self.query_feed(**kw)
    
    def reset(self, start=None, state=None):
        """
        resets internal model state.
        Args:
            start: if True, send start tokens through the model
                default behavior is True when state=None, False otherwise
            state: set the state from a result of `get_state`,
                instead of the initial state
        """
        self.step = 0
        if start is None:
            start = state is None
        if state is None: 
            named_states = zip(self.cell_state_names(), self.initial_state)
        else:
            named_states = state.items()
        with torch.inference_mode():
            for n,t in named_states:
                getattr(self, n)[:] = t
            if start:
                self.feed(
                    self.instrument_start_token, self.pitch_start_token, 0., 0.)
        # for n,t in zip(self.cell_state_names(), self.initial_state):
        #     getattr(self, n)[:] = t.detach()
        # if start:
        #     self.feed(
        #         self.instrument_start_token, self.pitch_start_token, 0., 0.)

    def get_state(self) -> Dict[str, torch.Tensor]:
        """return a dict of {str:Tensor} representing the model state"""
        return {n:getattr(self, n).clone() for n in self.cell_state_names()}
                

    @classmethod
    def user_data_dir(cls):
        return _user_data_dir()

    @classmethod
    def from_checkpoint(cls, path):
        """
        create a Notochord from a checkpoint file containing 
        hyperparameters and model weights.

        Args:
            path: file path to Notochord model
        """
        if path=="notochord-latest.ckpt":
            url = 'https://github.com/Intelligent-Instruments-Lab/iil-python-tools/releases/download/notochord-v0.4.0/notochord_lakh_50G_deep.pt'
        elif path=="txala-latest.ckpt":
            url = 'https://github.com/Intelligent-Instruments-Lab/notochord/releases/download/notochord-v0.5.4/noto-txala-011-0020.ckpt'
        else:
            url = None

        if url is not None:
            d = Notochord.user_data_dir()
            path = d / path
            # maybe download
            if not path.is_file():
                while True:
                    answer = input("Do you want to download a notochord model? (y/n)")
                    if answer.lower() in ["y","yes"]:
                        download_url(url, path)
                        print(f'saved to {path}')
                        break
                    if answer.lower() in ["n","no"]:
                        break
        # path = 
        checkpoint = torch.load(path, map_location=torch.device('cpu'))
        model = cls(**checkpoint['kw']['model'])
        model.load_state_dict(checkpoint['model_state'], strict=False)
        model.checkpoint_path = path
        return model
    
    def prompt(self, midi_file):
        """Read a MIDI file and feed events to this Notochord model.

        When possible, the hidden states will be cached so re-using the same prompt will be fast.

        Args:
            midi_file: path of a midi file to read
        Returns:
            state: hidden state dict of the Notochord encoding the MIDI prompt
            channel_inst: dict mapping MIDI channel (0-index) to Notochord instrument (1-256)
        """
        return prompt(self, midi_file, state_hash=hash_states(self.get_state()))
    
def hash_states(s):
    if isinstance(s, dict):
        return {k:hash_states(v) for k,v in s.items()}
    elif isinstance(s, torch.Tensor):
        return hash_tensor(s)
    return s

def hash_tensor(t):
    return hashlib.md5(json.dumps(t.tolist()).encode('utf-8')).digest()

@mem.cache(ignore=('noto',))
def prompt(noto:Notochord, midi_file:str|Path, state_hash:int|None=None): 
    # checkpoint name used for disk cache
    """Read a MIDI file and feed events to a Notochord model.

    Args:
        noto: a Notochord
        midi_file: path of a midi file to read
        state_hash: representation of model hidden state to use for caching results
    Returns:
        state: hidden state dict of the Notochord encoding the MIDI prompt
        channel_inst: dict mapping MIDI channel (0-index) to Notochord instrument (1-256)
    """
    # TODO: deduplicate this code?
    class AnonTracks:
        def __init__(self):
            self.n = 0
        def __call__(self):
            self.n += 1
            return 256+self.n
    next_anon = AnonTracks()
    mid_channel_inst = defaultdict(next_anon)

    mid = mido.MidiFile(midi_file)
    ticks_per_beat = mid.ticks_per_beat
    us_per_beat = 500_000
    time_seconds = 0
    prev_time_seconds = 0
    event_count = defaultdict(int)
    print(f'MIDI file: {ticks_per_beat} ticks, {us_per_beat} μs per beat')

    for msg in tqdm(mid, desc='ingesting MIDI prompt'):
        # when iterating over a track this is ticks,
        # when iterating the whole file it's seconds
        time_seconds += msg.time

        if msg.type=='program_change':
            mid_channel_inst[msg.channel] = msg.program + 1 + 128*int(msg.channel==9)
            tqdm.write(str(msg))
            
        elif msg.type=='set_tempo':
            us_per_beat = msg.tempo
            tqdm.write(f'MIDI file: set tempo {us_per_beat} μs/beat at {time_seconds} seconds')
            
        elif msg.type in ('note_on', 'note_off'):
            # make channel 10 with no PC standard drumkit
            if msg.channel not in mid_channel_inst and msg.channel==9:
                mid_channel_inst[msg.channel] = 129

            event_count[mid_channel_inst[msg.channel]] += 1
            noto.feed(
                mid_channel_inst[msg.channel],
                msg.note,
                time_seconds - prev_time_seconds,
                msg.velocity if msg.type=='note_on' else 0)
            prev_time_seconds = time_seconds

        else: continue

    print(f'MIDI file: {sum(event_count.values())} events in {time_seconds} seconds')
    # print(event_count)
    # print(mid_channel_inst)  
    return noto.get_state(), dict(mid_channel_inst)